from .channels import merge_channels, split_channels
from .consistency import image_physical_space_consistency, allclose
from .get_ants_data import get_ants_data, get_data
from .matrix_image import (matrix_to_images,
                           images_from_matrix,
                           image_list_to_matrix,
                           images_to_matrix,
                           matrix_from_images,
                           timeseries_to_matrix,
                           matrix_to_timeseries)
from .mni2tal import mni2tal
from .ndimage_to_list import ndimage_to_list, list_to_ndimage
from .nifti_to_ants import nifti_to_ants
from .scalar_rgb_vector import rgb_to_vector, vector_to_rgb, scalar_to_rgb
from .sitk_to_ants import from_sitk, to_sitk
from .convergence_monitoring import convergence_monitoring
from .polar_decomposition import polar_decomposition



"""
Test that any previously reported bugs are fixed

nptest.assert_allclose
self.assertEqual
self.assertTrue

"""

import os
import unittest
from common import run_tests

import ants

import numpy as np
import numpy.testing as nptest


class Test_bugs(unittest.TestCase):
    """
    Test ants.ANTsImage class
    """
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_resample_returns_NaNs(self):
        """
        Test that resampling an image doesnt cause the resampled
        image to have NaNs - previously caused by resampling an
        image of type DOUBLE
        """
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img2dr = ants.resample_image(img2d, (2,2), 0, 0)

        self.assertTrue(np.sum(np.isnan(img2dr.numpy())) == 0)

        img3d = ants.image_read(ants.get_ants_data('mni'))
        img3dr = ants.resample_image(img3d, (2,2,2), 0, 0)

        self.assertTrue(np.sum(np.isnan(img3dr.numpy())) == 0)

    def test_compose_multi_type_transforms(self):
        image = ants.image_read(ants.get_ants_data("r16"))

        linear_transform = ants.create_ants_transform(transform_type=
            "AffineTransform", precision='float', dimension=image.dimension)

        displacement_field = ants.simulate_displacement_field(image,
            field_type="bspline", number_of_random_points=1000,
            sd_noise=10.0, enforce_stationary_boundary=True,
            number_of_fitting_levels=4, mesh_size=1,
            sd_smoothing=4.0)
        displacement_field_xfrm = ants.transform_from_displacement_field(displacement_field)

        xfrm = ants.compose_ants_transforms([linear_transform, displacement_field_xfrm])
        xfrm = ants.compose_ants_transforms([linear_transform, linear_transform])
        xfrm = ants.compose_ants_transforms([displacement_field_xfrm, linear_transform])

    def test_bspline_image_with_2d_weights(self):
        # see https://github.com/ANTsX/ANTsPy/issues/655
        import ants
        import numpy as np

        output_size = (256, 256)
        bspline_epsilon = 1e-4
        number_of_fitting_levels = 4

        image = ants.image_read(ants.get_ants_data("r16"))
        image = ants.resample_image(image, (100, 100), use_voxels=True)

        indices = np.meshgrid(list(range(image.shape[0])), 
                                list(range(image.shape[1])))
        indices_array = np.stack((indices[1].flatten(), 
                                indices[0].flatten()), axis=0)

        image_parametric_values = indices_array.transpose()

        weight_array = np.ones(image.shape)
        parametric_values = image_parametric_values
        scattered_data = np.atleast_2d(image.numpy().flatten()).transpose()
        weight_values = np.atleast_2d(weight_array.flatten()).transpose()

        min_parametric_values = np.min(parametric_values, axis=0)
        max_parametric_values = np.max(parametric_values, axis=0)

        spacing = np.zeros((2,))
        for d in range(2):
            spacing[d] = (max_parametric_values[d] - min_parametric_values[d]) / (output_size[d] - 1) + bspline_epsilon

        bspline_image = ants.fit_bspline_object_to_scattered_data(scattered_data, parametric_values, 
                                                                parametric_domain_origin=min_parametric_values - bspline_epsilon,
                                                                parametric_domain_spacing=spacing,
                                                                parametric_domain_size=output_size,
                                                                data_weights=weight_values,
                                                                number_of_fitting_levels=number_of_fitting_levels,
                                                                mesh_size=1)


    def test_scalar_rgb_missing(self):
        import ants
        img = ants.image_read(ants.get_data('r16'))
        with self.assertRaises(Exception):
            img_color = ants.scalar_to_rgb(img, cmap='jet')

    def test_bspline_zeros(self):
        import ants
        import numpy as np
        x = np.linspace(-4, 4, num=100) + np.random.uniform(-0.1, 0.1, 100)
        u = np.linspace(0, 1.0, num=len(x))
        scattered_data = np.expand_dims(u, axis=-1)
        parametric_data = np.expand_dims(u, axis=-1)
        spacing = 1/(len(x)-1) * 1.0
        bspline_curve = ants.fit_bspline_object_to_scattered_data(
                scattered_data, parametric_data,
                parametric_domain_origin=[0.0], parametric_domain_spacing=[spacing],
                parametric_domain_size=[len(x)], is_parametric_dimension_closed=None,
                number_of_fitting_levels=5, mesh_size=1)

        # Erroneously returns all zeros.
        self.assertNotEqual(bspline_curve.sum(), 0)

    def test_from_numpy_different_dtypes(self):
        all_dtypes = ('bool',
                      'int8',
                      'int16',
                      'int32',
                      'int64',
                      'uint16',
                      'uint64',
                      'float16')
        arr = np.random.randn(100,100)
        
        for dtype in all_dtypes:
            arr2 = arr.astype(dtype)
            img = ants.from_numpy(arr2)
            self.assertTrue(ants.is_image(img))
        
if __name__ == '__main__':
    run_tests()



__all__ = ['mask_image']

import numpy as np

import ants
from ants.decorators import image_method

@image_method
def mask_image(image, mask, level=1, binarize=False):
    """
    Mask an input image by a mask image.  If the mask image has multiple labels,
    it is possible to specify which label(s) to mask at.

    ANTsR function: `maskImage`

    Arguments
    ---------
    image : ANTsImage
        Input image.

    mask : ANTsImage
        Mask or label image.

    level : scalar or tuple of scalars
        Level(s) at which to mask image. If vector or list of values, output image is non-zero at all locations where label image matches any of the levels specified.

    binarize : boolean
        whether binarize the output image

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> myimage = ants.image_read(ants.get_ants_data('r16'))
    >>> mask = ants.get_mask(myimage)
    >>> myimage_mask = ants.mask_image(myimage, mask, 3)
    >>> seg = ants.kmeans_segmentation(myimage, 3)
    >>> myimage_mask = ants.mask_image(myimage, seg['segmentation'], (1,3))
    """
    leveluse = level
    if type(leveluse) is np.ndarray:
        leveluse = level.tolist()
    if type(leveluse) is int or type(leveluse) is float:
        leveluse = [level]
    image_out = image.clone() * 0
    for mylevel in leveluse:
        temp = ants.threshold_image(mask, mylevel, mylevel)
        if binarize:
            image_out = image_out + temp
        else:
            image_out = image_out + temp * image
    return image_out




__all__ = ['prior_based_segmentation']

import ants


def prior_based_segmentation(image, priors, mask, priorweight=0.25, mrf=0.1, iterations=25):
    """
    Spatial prior-based image segmentation.
    
    Markov random field regularized, prior-based image segmentation that is a 
    wrapper around atropos (see ANTs and related publications).

    ANTsR function: `priorBasedSegmentation`

    Arguments
    ---------
    image : ANTsImage or list/tuple of ANTsImage types
        input image or image list for multivariate segmentation

    priors : list/tuple of ANTsImage types
        list of priors that cover the number of classes

    mask : ANTsImage
        segment inside this mask

    prior_weight : scalar
        usually 0 (priors used for initialization only), 0.25 or 0.5.

    mrf : scalar
        regularization, higher is smoother, a numerical value in range 0.0 to 0.2

    iterations : integer
        maximum number of iterations.  could be a large value eg 25.

    Returns
    -------
    dictionary with the following key/value pairs:
        `segmentation`: ANTsImage 
            actually segmented image
        
        `probabilityimages` : list of ANTsImage types
            one image for each segmentation class

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16'))
    >>> seg = ants.kmeans_segmentation(fi,3)
    >>> mask = ants.threshold_image(seg['segmentation'], 1, 1e15)
    >>> priorseg = ants.prior_based_segmentation(fi, seg['probabilityimages'], mask, 0.25, 0.1, 3)
    """
    if ants.is_image(image):
        dim = image.dimension
    elif isinstance(image, (tuple,list)) and (ants.is_image(image[0])):
        dim = image[0].dimension
    else:
        raise ValueError('image argument must be ANTsImage or list/tuple of ANTsImage types')

    nhood = 'x'.join(['1']*dim)
    mrf = '[%s,%s]' % (str(mrf), nhood)
    conv = '[%s,0]' % (str(iterations))

    pseg = ants.atropos(a=image, m=mrf, c=conv, i=priors, x=mask, priorweight=priorweight)

    return pseg





from .averaging import average_images
from .get_centroids import get_centroids
from .get_neighborhood import get_neighborhood_in_mask, get_neighborhood_at_voxel
from .hausdorff_distance import hausdorff_distance
from .image_similarity import image_similarity
from .metrics import image_mutual_information
from .quantile import (ilr,
                       rank_intensity,
                       quantile,
                       regress_poly,
                       regress_components,
                       get_average_of_timeseries,
                       compcor,
                       bandpass_filter_matrix)


__all__ = ["fit_bspline_object_to_scattered_data"]

import numpy as np

import ants
from ants.internal import get_lib_fn


def fit_bspline_object_to_scattered_data(scattered_data,
                                         parametric_data,
                                         parametric_domain_origin,
                                         parametric_domain_spacing,
                                         parametric_domain_size,
                                         is_parametric_dimension_closed=None,
                                         data_weights=None,
                                         number_of_fitting_levels=4,
                                         mesh_size=1,
                                         spline_order=3):

    """
    Fit a b-spline object to scattered data.  This is basically a wrapper
    for the ITK filter

    https://itk.org/Doxygen/html/classitk_1_1BSplineScatteredDataPointSetToImageFilter.html

    This filter is flexible in the possible objects that can be approximated.
    Possibilities include:

        * 1/2/3/4-D curve
        * 2-D surface in 3-D space (not available/templated)
        * 2/3/4-D scalar field
        * 2/3-D displacement field
        * 2/3-D time-varying velocity field

    In order to understand the input parameters, it is important to understand
    the difference between the parametric and data dimensions.  A curve as one
    parametric dimension but the data dimension can be 1-D, 2-D, 3-D, or 4-D.
    In contrast, a 3-D displacement field has a parametric and data dimension
    of 3.  The scattered data is what's approximated by the B-spline object and
    the parametric point is the location of scattered data within the domain of
    the B-spline object.

    ANTsR function: `fitBsplineObjectToScatteredData`

    Arguments
    ---------
    scattered_data : 2-D numpy array
        Defines the scattered data input to be approximated. Data is organized
        by row --> data v, column ---> data dimension.

    parametric_data : 2-D numpy array
        Defines the parametric location of the scattered data.  Data is organized
        by row --> parametric point, column --> parametric dimension.  Note that
        each row corresponds to the same row in the scatteredData.

    data_weights : 1-D numpy array
        Defines the individual weighting of the corresponding scattered data value.
        Default = None meaning all values are weighted the same.

    parametric_domain_origin : n-D tuple
        Defines the parametric origin of the B-spline object.

    parametric_domain_spacing : n-D tuple
        Defines the parametric spacing of the B-spline object.  Defines the sampling
        rate in the parametric domain.

    parametric_domain_size : n-D tuple
       Defines the size (length) of the B-spline object.  Note that the length of the
       B-spline object in dimension d is defined as
       parametric_domain_spacing[d] * parametric_domain_size[d]-1.

    is_parametric_dimension_closed : n-D tuple
       Booleans defining whether or not the corresponding parametric dimension is
       closed (e.g., closed loop).  Default = None.

    number_of_fitting_levels : integer
       Specifies the number of fitting levels.

    mesh_size : n-D tuple
       Defines the mesh size at the initial fitting level.

    spline_order : integer
       Spline order of the B-spline object.  Default = 3.

    Returns
    -------
    returns numpy array for B-spline curve (parametric dimension = 1).  Otherwise,
    returns an ANTsImage.

    Example
    -------
    >>> # Perform 2-D curve example
    >>>
    >>> import ants, numpy
    >>> import matplotlib.pyplot as plt
    >>> x = numpy.linspace(-4, 4, num=100)
    >>> y = numpy.exp(-numpy.multiply(x, x)) + numpy.random.uniform(-0.1, 0.1, len(x))
    >>> u = numpy.linspace(0, 1.0, num=len(x))
    >>> scattered_data = numpy.column_stack((x, y))
    >>> parametric_data = numpy.expand_dims(u, axis=-1)
    >>> spacing = 1/(len(x)-1) * 1.0;
    >>> bspline_curve = ants.fit_bspline_object_to_scattered_data(scattered_data, 
    >>>    parametric_data,
    >>>   parametric_domain_origin=[0.0], parametric_domain_spacing=[spacing],
    >>>   parametric_domain_size=[len(x)], is_parametric_dimension_closed=None,
    >>>   number_of_fitting_levels=5, mesh_size=1)
    >>> plt.plot(x, y, label='Noisy points')
    >>> plt.plot(bspline_curve[:,0], bspline_curve[:,1], label='B-spline curve')
    >>> plt.grid(True)
    >>> plt.axis('tight')
    >>> plt.legend(loc='upper left')
    >>> plt.show()
    >>>
    >>> ###########################################################################
    >>>
    >>> # Perform 2-D scalar field (i.e., image) example
    >>>
    >>> import ants, numpy
    >>> number_of_random_points = 10000
    >>> img = ants.image_read( ants.get_ants_data("r16"))
    >>> img_array = img.numpy()
    >>> row_indices = numpy.random.choice(range(2, img_array.shape[0]), number_of_random_points)
    >>> col_indices = numpy.random.choice(range(2, img_array.shape[1]), number_of_random_points)
    >>> scattered_data = numpy.zeros((number_of_random_points, 1))
    >>> parametric_data = numpy.zeros((number_of_random_points, 2))
    >>> for i in range(number_of_random_points):
    >>>     scattered_data[i,0] = img_array[row_indices[i], col_indices[i]]
    >>>     parametric_data[i,0] = row_indices[i]
    >>>     parametric_data[i,1] = col_indices[i]
    >>> bspline_img = ants.fit_bspline_object_to_scattered_data(
    >>>     scattered_data, parametric_data,
    >>>     parametric_domain_origin=[0.0, 0.0],
    >>>     parametric_domain_spacing=[1.0, 1.0],
    >>>     parametric_domain_size = img.shape,
    >>>     number_of_fitting_levels=7, mesh_size=1)
    >>>
    >>> ants.plot(img, title="Original")
    >>> ants.plot(bspline_img, title="B-spline approximation")
    """

    parametric_dimension = parametric_data.shape[1]
    data_dimension = scattered_data.shape[1]

    if is_parametric_dimension_closed is None:
        is_parametric_dimension_closed = np.repeat(False, parametric_dimension)

    if isinstance(is_parametric_dimension_closed, bool):
        is_parametric_dimension_closed = np.repeat(is_parametric_dimension_closed, parametric_dimension)    

    if len(is_parametric_dimension_closed) != parametric_dimension:
        raise ValueError("Closed is not of length parametric_dimension.")

    if isinstance(mesh_size, int) == False and len(mesh_size) != parametric_dimension:
        raise ValueError("Incorrect specification for mesh_size.")

    if isinstance(mesh_size, int):
        number_of_control_points = np.repeat(mesh_size + spline_order, parametric_dimension)
    else:
        number_of_control_points = np.array(mesh_size) + np.repeat(spline_order, parametric_dimension)

    for i in range(len(is_parametric_dimension_closed)):
        if is_parametric_dimension_closed[i]:
            if number_of_control_points[i] < 2 * spline_order + 1:
                raise ValueError("For closed parametric dimension, mesh_size > spline_order")

    if len(parametric_domain_origin) != parametric_dimension:
        raise ValueError("Origin is not of length parametric_dimension.")

    if len(parametric_domain_spacing) != parametric_dimension:
        raise ValueError("Spacing is not of length parametric_dimension.")

    if len(parametric_domain_size) != parametric_dimension:
        raise ValueError("Size is not of length parametric_dimension.")    

    if parametric_data.shape[0] != scattered_data.shape[0]:
        raise ValueError("The number of points is not equal to the number of scattered data values.")

    if data_weights is None:
        data_weights = np.repeat(1.0, parametric_data.shape[0])
    
    if data_weights.ndim == 2:
        data_weights = np.squeeze(data_weights)

    if len(data_weights) != parametric_data.shape[0]:
        raise ValueError("The number of weights is not the same as the number of points.")

    libfn = get_lib_fn("fitBsplineObjectToScatteredDataP%iD%i" % (parametric_dimension, data_dimension))
    bspline_object = libfn(scattered_data, parametric_data, np.array(data_weights).tolist(),
                           np.array(parametric_domain_origin).tolist(), np.array(parametric_domain_spacing).tolist(),
                           np.array(parametric_domain_size).tolist(), np.array(is_parametric_dimension_closed).tolist(),
                           number_of_fitting_levels, np.array(number_of_control_points).tolist(),
                           spline_order)

    if parametric_dimension == 1:
        return np.array(bspline_object)
    else:
        bspline_image = ants.from_pointer(bspline_object).clone('float')
        return bspline_image




"""
"""
__all__ = ['new_ants_metric',
           'create_ants_metric',
           'supported_metrics']


import os
import numpy as np

import ants
from ants.internal import get_lib_fn
from ants.core.ants_metric import ANTsImageToImageMetric

_supported_metrics = {'MeanSquares',
                    'MattesMutualInformation',
                    'ANTSNeighborhoodCorrelation',
                    'Correlation',
                    'Demons',
                    'JointHistogramMutualInformation'}


def supported_metrics():
    return _supported_metrics

def new_ants_metric(dimension=3, precision='float', metric_type='MeanSquares'):
    if metric_type not in _supported_metrics:
        raise ValueError('metric_type must be one of %s' % _supported_metrics)

    libfn = get_lib_fn('new_ants_metricF%i'%dimension)
    itk_tx = libfn(precision, dimension, metric_type)

    ants_metric = ANTsImageToImageMetric(itk_tx)
    return ants_metric


def create_ants_metric(fixed,
                        moving,
                        metric_type='MeanSquares',
                        fixed_mask=None,
                        moving_mask=None,
                        sampling_strategy='regular',
                        sampling_percentage=1):
    """
    Arguments
    ---------
    metric_type : string
        which metric to use
        options:
            MeanSquares
            MattesMutualInformation
            ANTSNeighborhoodCorrelation
            Correlation
            Demons
            JointHistogramMutualInformation

    Example
    -------
    >>> import ants
    >>> fixed = ants.image_read(ants.get_ants_data('r16'))
    >>> moving = ants.image_read(ants.get_ants_data('r64'))
    >>> metric_type = 'Correlation'
    >>> metric = ants.create_ants_metric(fixed, moving, metric_type)
    """
    dimension = fixed.dimension
    pixeltype = fixed.pixeltype
    is_vector = False # For now, no multichannel images

    if metric_type not in _supported_metrics:
        raise ValueError('metric_type must be one of %s' % _supported_metrics)

    if (dimension < 2) or (dimension > 4):
        raise ValueError('unsupported dimension %i' % dimension)

    if not ants.is_image(moving):
        raise ValueError('invalid moving image')

    if moving.dimension != dimension:
        raise ValueError('Fixed and Moving images must have same dimension')

    if not ants.is_image(fixed):
        raise ValueError('invalid fixed image')

    fixed = fixed.clone('float')
    moving = moving.clone('float')

    libfn = get_lib_fn('create_ants_metricF%i' % dimension)
    metric = libfn(pixeltype, dimension, metric_type, is_vector, fixed.pointer, moving.pointer)

    ants_metric = ANTsImageToImageMetric(metric)
    ants_metric.set_fixed_image(fixed)
    ants_metric.set_moving_image(moving)

    if ants.is_image(fixed_mask):
        ants_metric.set_fixed_mask(fixed_mask)

    if ants.is_image(moving_mask):
        ants_metric.set_moving_mask(moving_mask)

    ants_metric.set_sampling(sampling_strategy, sampling_percentage)
    ants_metric.initialize()

    return ants_metric





 

__all__ = ['merge_channels',
           'split_channels']




import ants
from ants.internal import get_lib_fn
from ants.decorators import image_method


def merge_channels(image_list, channels_first=False):
    """
    Merge channels of multiple scalar ANTsImage types into one 
    multi-channel ANTsImage
    
    ANTsR function: `mergeChannels`

    Arguments
    ---------
    image_list : list/tuple of ANTsImage types
        scalar images to merge
    
    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image2 = ants.image_read(ants.get_ants_data('r16'))
    >>> image3 = ants.merge_channels([image,image2])
    >>> image3 = ants.merge_channels([image,image2], channels_first=True)
    >>> image3.numpy()
    >>> image3.components == 2
    """
    inpixeltype = image_list[0].pixeltype
    dimension = image_list[0].dimension
    components = len(image_list)

    for image in image_list:
        if not ants.is_image(image):
            raise ValueError('list may only contain ANTsImage objects')
        if image.pixeltype != inpixeltype:
            raise ValueError('all images must have the same pixeltype')

    libfn = get_lib_fn('mergeChannels')
    image_ptr = libfn([image.pointer for image in image_list])
    
    image = ants.from_pointer(image_ptr)
    image.channels_first = channels_first
    return image

@image_method
def split_channels(image):
    """
    Split channels of a multi-channel ANTsImage into a collection
    of scalar ANTsImage types
    
    Arguments
    ---------
    image : ANTsImage
        multi-channel image to split

    Returns
    -------
    list of ANTsImage types

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'), 'float')
    >>> image2 = ants.image_read(ants.get_ants_data('r16'), 'float')
    >>> imagemerge = ants.merge_channels([image,image2])
    >>> imagemerge.components == 2
    >>> images_unmerged = ants.split_channels(imagemerge)
    >>> len(images_unmerged) == 2
    >>> images_unmerged[0].components == 1
    """
    inpixeltype = image.pixeltype
    dimension = image.dimension
    components = 1

    libfn = get_lib_fn('splitChannels')
    itkimages = libfn(image.pointer)
    antsimages = [ants.from_pointer(itkimage) for itkimage in itkimages]
    return antsimages







import sys
import os
import argparse
import unittest
import warnings
import contextlib
from functools import wraps
from itertools import product
from copy import deepcopy


def run_tests():
    import faulthandler
    faulthandler.enable()
    unittest.main(verbosity=2)




__all__ = ['label_image_centroids']

import numpy as np

from ants.decorators import image_method


@image_method
def label_image_centroids(image, physical=False, convex=True, verbose=False):
    """
    Converts a label image to coordinates summarizing their positions

    ANTsR function: `labelImageCentroids`

    Arguments
    ---------
    image : ANTsImage
        image of integer labels

    physical : boolean
        whether you want physical space coordinates or not

    convex : boolean
        if True, return centroid
        if False return point with min average distance to other points with same label

    Returns
    -------
    dictionary w/ following key-value pairs:
        `labels` : 1D-ndarray
            array of label values

        `vertices` : pd.DataFrame
            coordinates of label centroids

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> image = ants.from_numpy(np.asarray([[[0,2],[1,3]],[[4,6],[5,7]]]).astype('float32'))
    >>> labels = ants.label_image_centroids(image)
    """
    d = image.shape
    if len(d) != 3:
        raise ValueError('image must be 3 dimensions')

    xcoords = np.asarray(np.arange(d[0]).tolist()*(d[1]*d[2]))
    ycoords = np.asarray(np.repeat(np.arange(d[1]),d[0]).tolist()*d[2])
    zcoords = np.asarray(np.repeat(np.arange(d[1]), d[0]*d[2]))

    labels = image.numpy()
    mylabels = np.sort(np.unique(labels[labels > 0])).astype('int')
    n_labels = len(mylabels)
    xc = np.zeros(n_labels)
    yc = np.zeros(n_labels)
    zc = np.zeros(n_labels)

    if convex:
        for lab_idx, label_intensity in enumerate(mylabels):
            idx = (labels == label_intensity).flatten()
            xc[lab_idx] = np.mean(xcoords[idx])
            yc[lab_idx] = np.mean(ycoords[idx])
            zc[lab_idx] = np.mean(zcoords[idx])
    else:
        for lab_idx, label_intensity in enumerate(mylabels):
            idx = (labels == label_intensity).flatten()
            xci = xcoords[idx]
            yci = ycoords[idx]
            zci = zcoords[idx]
            dist = np.zeros(len(xci))

            for j in range(len(xci)):
                dist[j] = np.mean(np.sqrt((xci[j] - xci)**2 + (yci[j] - yci)**2 + (zci[j] - zci)**2))

            mid = np.where(dist==np.min(dist))
            xc[lab_idx] = xci[mid]
            yc[lab_idx] = yci[mid]
            zc[lab_idx] = zci[mid]

    centroids = np.vstack([xc,yc,zc]).T

    #if physical:
    #    centroids = ants.transform_index_to_physical_point(image, centroids)

    return {
        'labels': mylabels,
        'vertices': centroids
    }









__all__ = ['create_warped_grid']

import numpy as np

import ants


def create_warped_grid(image, grid_step=10, grid_width=2, grid_directions=(True, True),
                       fixed_reference_image=None, transform=None, foreground=1, background=0):
    """
    Deforming a grid is a helpful way to visualize a deformation field. 
    This function enables a user to define the grid parameters 
    and apply a deformable map to that grid.

    ANTsR function: `createWarpedGrid`
    
    Arguments
    ---------
    image : ANTsImage
        input image
    
    grid_step : scalar   
        width of grid blocks
    
    grid_width : scalar
        width of grid lines
    
    grid_directions : tuple of booleans 
        directions in which to draw grid lines, boolean vector
    
    fixed_reference_image : ANTsImage (optional)
        reference image space
    
    transform : list/tuple of strings (optional)
        vector of transforms
    
    foreground : scalar
        intensity value for grid blocks
    
    background : scalar
        intensity value for grid lines

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data( 'r16' ) )
    >>> mi = ants.image_read( ants.get_ants_data( 'r64' ) )
    >>> mygr = ants.create_warped_grid( mi )
    >>> mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = ('SyN') )
    >>> mywarpedgrid = ants.create_warped_grid( mygr, grid_directions=(False,True),
                            transform=mytx['fwdtransforms'], fixed_reference_image=fi )
    """
    if ants.is_image(image):
        if len(grid_directions) != image.dimension:
            grid_directions = [True]*image.dimension
        garr = image.numpy() * 0 + foreground
    else:
        if not isinstance(image, (list, tuple)):
            raise ValueError('image arg must be ANTsImage or list or tuple')
        if len(grid_directions) != len(image):
            grid_directions = [True]*len(image)
        garr = np.zeros(image) + foreground
        image = ants.from_numpy(garr)

    idim = garr.ndim
    gridw = grid_width

    for d in range(idim):
        togrid = np.arange(-1, garr.shape[d]-1, step=grid_step)
        for i in range(len(togrid)):
            if (d == 0) & (idim == 3) & (grid_directions[d]):
                garr[togrid[i]:(togrid[i]+gridw),...] = background
                garr[0,...] = background
                garr[-1,...] = background
            if (d == 1) & (idim == 3) & (grid_directions[d]):
                garr[:,togrid[i]:(togrid[i]+gridw),:] = background
                garr[:,0,:] = background
                garr[:,-1,:] = background
            if (d == 2) & (idim == 3) & (grid_directions[d]):
                garr[...,togrid[i]:(togrid[i]+gridw)] = background
                garr[...,0] = background
                garr[...,-1] = background
            if (d == 0) & (idim == 2) & (grid_directions[d]):
                garr[togrid[i]:(togrid[i]+gridw),:] = background
                garr[0,:] = background
                garr[-1,:] = background
            if (d == 1) & (idim == 2) & (grid_directions[d]):
                garr[:,togrid[i]:(togrid[i]+gridw)] = background
                garr[:,0] = background
                garr[:,-1] = background


    gimage = image.new_image_like(garr)

    if (transform is not None) and (fixed_reference_image is not None):
        return ants.apply_transforms( fixed=fixed_reference_image, moving=gimage,
                               transformlist=transform ) 
    else:
        return gimage



"""
Image IO
"""

__all__ = [
    "image_header_info",
    "image_clone",
    "image_read",
    "dicom_read",
    "image_write",
    "make_image",
    "from_numpy",
    "from_numpy_like",
    "new_image_like"
]

import os
import json
import numpy as np
import warnings

import ants
from ants.internal import get_lib_fn, short_ptype, infer_dtype
from ants.decorators import image_method

_supported_pclasses = {"scalar", "vector", "rgb", "rgba","symmetric_second_rank_tensor"}
_supported_ptypes = {"unsigned char", "unsigned int", "float", "double"}
_supported_ntypes = {"uint8", "uint32", "float32", "float64"}
_unsupported_ptypes = {"char", "unsigned short", "short", "int"}
_unsupported_ptype_map = {
    "char": "float",
    "unsigned short": "unsigned int",
    "short": "float",
    "int": "float",
}
_image_type_map = {"scalar": "", "vector": "V", "rgb": "RGB", "rgba": "RGBA", "symmetric_second_rank_tensor": "SSRT" }
_ptype_type_map = {
    "unsigned char": "UC",
    "unsigned int": "UI",
    "float": "F",
    "double": "D",
}

_ntype_type_map = {"uint8": "UC", "uint32": "UI", "float32": "F", "float64": "D"}
_npy_to_itk_map = {
    "uint8": "unsigned char",
    "uint32": "unsigned int",
    "float32": "float",
    "float64": "double",
}

_image_read_dict = {}
for itype in {"scalar", "vector", "rgb", "rgba", "symmetric_second_rank_tensor"}:
    _image_read_dict[itype] = {}
    for p in _supported_ptypes:
        _image_read_dict[itype][p] = {}
        for d in {2, 3, 4}:
            ita = _image_type_map[itype]
            pa = _ptype_type_map[p]
            _image_read_dict[itype][p][d] = "imageRead%s%s%i" % (ita, pa, d)

def from_numpy(
    data, origin=None, spacing=None, direction=None, has_components=False, is_rgb=False
):
    """
    Create an ANTsImage object from a numpy array

    ANTsR function: `as.antsImage`

    Arguments
    ---------
    data : ndarray
        image data array

    origin : tuple/list
        image origin

    spacing : tuple/list
        image spacing

    direction : list/ndarray
        image direction

    has_components : boolean
        whether the image has components

    Returns
    -------
    ANTsImage
        image with given data and any given information
    """

    # this is historic but should be removed once tests can pass without it
    if data.dtype.name == 'float64':
        data = data.astype('float32')

    # if dtype is not supported, cast to best available
    best_dtype = infer_dtype(data.dtype)
    if best_dtype != data.dtype:
        data = data.astype(best_dtype)

    # Be explicit about ordering of data - needs to be C-contiguous
    img = _from_numpy(data.T.copy(order='C'), origin, spacing, direction, has_components, is_rgb)
    return img


def _from_numpy(
    data, origin=None, spacing=None, direction=None, has_components=False, is_rgb=False
):
    """
    Internal function for creating an ANTsImage
    """
    if is_rgb:
        has_components = True
    ndim = data.ndim
    if has_components:
        ndim -= 1
    dtype = data.dtype.name
    ptype = _npy_to_itk_map[dtype]

    data = np.array(data)

    if origin is None:
        origin = tuple([0.0] * ndim)
    if spacing is None:
        spacing = tuple([1.0] * ndim)
    if direction is None:
        direction = np.eye(ndim)

    libfn = get_lib_fn("fromNumpy%s%i" % (_ntype_type_map[dtype], ndim))

    if not has_components:
        itk_image = libfn(data, data.shape[::-1])
        ants_image = ants.from_pointer(itk_image)
        ants_image.set_origin(origin)
        ants_image.set_spacing(spacing)
        ants_image.set_direction(direction)
        ants_image._ndarr = data
    else:
        arrays = [data[i, ...].copy() for i in range(data.shape[0])]
        data_shape = arrays[0].shape
        ants_images = []
        for i in range(len(arrays)):
            tmp_ptr = libfn(arrays[i], data_shape[::-1])
            tmp_img = ants.from_pointer(tmp_ptr)
            tmp_img.set_origin(origin)
            tmp_img.set_spacing(spacing)
            tmp_img.set_direction(direction)
            tmp_img._ndarr = arrays[i]
            ants_images.append(tmp_img)
        ants_image = ants.merge_channels(ants_images)
        if is_rgb:
            ants_image = ants_image.vector_to_rgb()
    return ants_image


def make_image(
    imagesize,
    voxval=0,
    spacing=None,
    origin=None,
    direction=None,
    has_components=False,
    pixeltype="float",
):
    """
    Make an image with given size and voxel value or given a mask and vector

    ANTsR function: `makeImage`

    Arguments
    ---------
    shape : tuple/ANTsImage
        input image size or mask

    voxval : scalar
        input image value or vector, size of mask

    spacing : tuple/list
        image spatial resolution

    origin  : tuple/list
        image spatial origin

    direction : list/ndarray
        direction matrix to convert from index to physical space

    components : boolean
        whether there are components per pixel or not

    pixeltype : float
        data type of image values

    Returns
    -------
    ANTsImage
    """
    if ants.is_image(imagesize):
        img = imagesize.clone()
        sel = imagesize > 0
        if voxval.ndim > 1:
            voxval = voxval.flatten()
        if (len(voxval) == int((sel > 0).sum())) or (len(voxval) == 0):
            img[sel] = voxval
        else:
            raise ValueError(
                "Num given voxels %i not same as num positive values %i in `imagesize`"
                % (len(voxval), int((sel > 0).sum()))
            )
        return img
    else:
        if isinstance(voxval, (tuple, list, np.ndarray)):
            array = np.asarray(voxval).astype("float32").reshape(imagesize)
        else:
            array = np.full(imagesize, voxval, dtype="float32")
        image = from_numpy(
            array,
            origin=origin,
            spacing=spacing,
            direction=direction,
            has_components=has_components,
        )
        return image.clone(pixeltype)


def image_header_info(filename):
    """
    Read file info from image header

    ANTsR function: `antsImageHeaderInfo`

    Arguments
    ---------
    filename : string
        name of image file from which info will be read

    Returns
    -------
    dict
    """
    if not os.path.exists(filename):
        raise Exception("filename does not exist")

    libfn = get_lib_fn("antsImageHeaderInfo")
    retval = libfn(filename)
    retval["dimensions"] = tuple(retval["dimensions"])
    retval["origin"] = tuple([round(o, 4) for o in retval["origin"]])
    retval["spacing"] = tuple([round(s, 4) for s in retval["spacing"]])
    retval["direction"] = np.round(retval["direction"], 4)
    return retval

def image_clone(image, pixeltype=None):
    """
    Clone an ANTsImage

    ANTsR function: `antsImageClone`

    Arguments
    ---------
    image : ANTsImage
        image to clone

    pixeltype : string (optional)
        new datatype for image

    Returns
    -------
    ANTsImage
    """
    return image.clone(pixeltype)


def image_read(filename, dimension=None, pixeltype="float", reorient=False):
    """
    Read an ANTsImage from file

    ANTsR function: `antsImageRead`

    Arguments
    ---------
    filename : string
        Name of the file to read the image from.

    dimension : int
        Number of dimensions of the image read. This need not be the same as
        the dimensions of the image in the file. Allowed values: 2, 3, 4.
        If not provided, the dimension is obtained from the image file

    pixeltype : string
        C++ datatype to be used to represent the pixels read. This datatype
        need not be the same as the datatype used in the file.
        Options: unsigned char, unsigned int, float, double

    reorient : boolean | string
        if True, the image will be reoriented to RPI if it is 3D
        if False, nothing will happen
        if string, this should be the 3-letter orientation to which the
            input image will reoriented if 3D.
        if the image is 2D, this argument is ignored

    Returns
    -------
    ANTsImage
    """
    if filename.endswith(".npy"):
        filename = os.path.expanduser(filename)
        img_array = np.load(filename)
        if os.path.exists(filename.replace(".npy", ".json")):
            with open(filename.replace(".npy", ".json")) as json_data:
                img_header = json.load(json_data)
            ants_image = from_numpy(
                img_array,
                origin=img_header.get("origin", None),
                spacing=img_header.get("spacing", None),
                direction=np.asarray(img_header.get("direction", None)),
                has_components=img_header.get("components", 1) > 1,
            )
        else:
            img_header = {}
            ants_image = from_numpy(img_array)

    else:
        filename = os.path.expanduser(filename)
        if not os.path.exists(filename):
            raise ValueError("File %s does not exist!" % filename)

        hinfo = image_header_info(filename)
        ptype = hinfo["pixeltype"]
        pclass = hinfo["pixelclass"]
        ndim = hinfo["nDimensions"]
        ncomp = hinfo["nComponents"]
        is_rgb = False
        if pclass == "rgb":
            pclass = "vector"
        if pclass == "rgba":
            pclass = "vector"
        if pclass == "symmetric_second_rank_tensor":
            pclass = "vector"
#        is_rgb = True if pclass == "rgb" else False
        if dimension is not None:
            ndim = dimension

        # error handling on pixelclass
        if pclass not in _supported_pclasses:
            raise ValueError("Pixel class %s not supported!" % pclass)

        # error handling on pixeltype
        if ptype in _unsupported_ptypes:
            ptype = _unsupported_ptype_map.get(ptype, "unsupported")
            if ptype == "unsupported":
                raise ValueError("Pixeltype %s not supported" % ptype)

        # error handling on dimension
        if (ndim < 2) or (ndim > 4):
            raise ValueError("Found %i-dimensional image - not supported!" % ndim)

        libfn = get_lib_fn(_image_read_dict[pclass][ptype][ndim])
        itk_pointer = libfn(filename)

        ants_image = ants.from_pointer(itk_pointer)

        if pixeltype is not None:
            ants_image = ants_image.clone(pixeltype)

    if (reorient != False) and (ants_image.dimension == 3):
        if reorient == True:
            ants_image = ants_image.reorient_image2("RPI")
        elif isinstance(reorient, str):
            ants_image = ants_image.reorient_image2(reorient)

    return ants_image


def dicom_read(directory, pixeltype="float"):
    """
    Read a set of dicom files in a directory into a single ANTsImage.
    The origin of the resulting 3D image will be the origin of the
    first dicom image read.

    Arguments
    ---------
    directory : string
        folder in which all the dicom images exist

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> img = ants.dicom_read('~/desktop/dicom-subject/')
    """
    slices = []
    imgidx = 0
    for imgpath in os.listdir(directory):
        if imgpath.endswith(".dcm"):
            if imgidx == 0:
                tmp = image_read(
                    os.path.join(directory, imgpath), dimension=3, pixeltype=pixeltype
                )
                origin = tmp.origin
                spacing = tmp.spacing
                direction = tmp.direction
                tmp = tmp.numpy()[:, :, 0]
            else:
                tmp = image_read(
                    os.path.join(directory, imgpath), dimension=2, pixeltype=pixeltype
                ).numpy()

            slices.append(tmp)
            imgidx += 1

    slices = np.stack(slices, axis=-1)
    return from_numpy(slices, origin=origin, spacing=spacing, direction=direction)

@image_method
def image_write(image, filename, ri=False):
    """
    Write an ANTsImage to file

    ANTsR function: `antsImageWrite`

    Arguments
    ---------
    image : ANTsImage
        image to save to file

    filename : string
        name of file to which image will be saved

    ri : boolean
        if True, return image. This allows for using this function in a pipeline:
            >>> img2 = img.smooth_image(2.).image_write(file1, ri=True).threshold_image(0,20).image_write(file2, ri=True)
        if False, do not return image
    """
    if filename.endswith(".npy"):
        img_array = image.numpy()
        img_header = {
            "origin": image.origin,
            "spacing": image.spacing,
            "direction": image.direction.tolist(),
            "components": image.components,
        }

        np.save(filename, img_array)
        with open(filename.replace(".npy", ".json"), "w") as outfile:
            json.dump(img_header, outfile)
    else:
        image.to_file(filename)

    if ri:
        return image

@image_method
def clone(image, pixeltype=None):
    """
    Create a copy of the given ANTsImage with the same data and info, possibly with
    a different data type for the image data. Only supports casting to
    uint8 (unsigned char), uint32 (unsigned int), float32 (float), and float64 (double)

    Arguments
    ---------
    pixeltype: string (optional)
        if None, the pixeltype will be the same as the cloned ANTsImage. Otherwise,
        the data will be cast to this type. This can be a numpy type or an ITK
        type.
        Options:
            'unsigned char' or 'uint8',
            'unsigned int' or 'uint32',
            'float' or 'float32',
            'double' or 'float64'

    Returns
    -------
    ANTsImage
    """
    if pixeltype is None:
        pixeltype = image.pixeltype

    if pixeltype not in _supported_ptypes:
        # check if the pixeltype is a numpy type
        if pixeltype in _supported_ntypes:
            pixeltype =  _npy_to_itk_map[pixeltype]
        else:
            raise ValueError('Pixeltype %s not supported. Supported types are %s' % (pixeltype, _supported_ptypes))

    if image.has_components and (not image.is_rgb):
        comp_imgs = ants.split_channels(image)
        comp_imgs_cloned = [comp_img.clone(pixeltype) for comp_img in comp_imgs]
        return ants.merge_channels(comp_imgs_cloned, channels_first=image.channels_first)
    else:
        p1_short = short_ptype(image.pixeltype)
        p2_short = short_ptype(pixeltype)
        ndim = image.dimension
        fn_suffix = '%s%i' % (p2_short,ndim)
        libfn = get_lib_fn('antsImageClone%s'%fn_suffix)
        pointer_cloned = libfn(image.pointer)
        return ants.from_pointer(pointer_cloned)

copy = clone

@image_method
def new_image_like(image, data):
    """
    Create a new ANTsImage with the same header information, but with
    a new image array.

    Arguments
    ---------
    data : ndarray or py::capsule
        New array or pointer for the image.
        It must have the same shape as the current
        image data.

    Returns
    -------
    ANTsImage
    """
    if not isinstance(data, np.ndarray):
        raise ValueError('data must be a numpy array')
    if not image.has_components:
        if data.shape != image.shape:
            raise ValueError('given array shape (%s) and image array shape (%s) do not match' % (data.shape, image.shape))
    else:
        if (data.shape[-1] != image.components) or (data.shape[:-1] != image.shape):
            raise ValueError('given array shape (%s) and image array shape (%s) do not match' % (data.shape[1:], image.shape))

    return from_numpy(data, origin=image.origin,
        spacing=image.spacing, direction=image.direction,
        has_components=image.has_components)

def from_numpy_like(data, image):
    return new_image_like(image, data)



__all__ = ['affine_initializer']

import warnings
from tempfile import mktemp

from ants.internal import get_lib_fn, process_arguments


def affine_initializer(fixed_image, moving_image, search_factor=20,
                        radian_fraction=0.1, use_principal_axis=False,
                        local_search_iterations=10, mask=None, txfn=None ):
    """
    A multi-start optimizer for affine registration
    Searches over the sphere to find a good initialization for further
    registration refinement, if needed.  This is a wrapper for the ANTs
    function antsAffineInitializer.

    ANTsR function: `affineInitializer`

    Arguments
    ---------
    fixed_image : ANTsImage
        the fixed reference image
    moving_image : ANTsImage
        the moving image to be mapped to the fixed space
    search_factor : scalar
        degree of increments on the sphere to search
    radian_fraction : scalar
        between zero and one, defines the arc to search over
    use_principal_axis : boolean
        boolean to initialize by principal axis
    local_search_iterations : scalar
        gradient descent iterations
    mask : ANTsImage (optional)
        optional mask to restrict registration
    txfn : string (optional)
        filename for the transformation

    Returns
    -------
    ndarray
        transformation matrix

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16'))
    >>> mi = ants.image_read(ants.get_ants_data('r27'))
    >>> txfile = ants.affine_initializer( fi, mi )
    >>> tx = ants.read_transform(txfile, dimension=2)
    """

    if txfn is None:
        txfn = mktemp(suffix='.mat')

    veccer = [fixed_image.dimension, fixed_image, moving_image, txfn,
                search_factor, radian_fraction, int(use_principal_axis),
                local_search_iterations]
    if mask is not None:
        veccer.append(mask)

    xxx = process_arguments(veccer)
    libfn = get_lib_fn('antsAffineInitializer')
    retval = libfn(xxx)

    if retval != 0:
        warnings.warn('ERROR: Non-zero exit status!')

    return txfn


import numpy as np
import ants

__all__ = ['polar_decomposition']

def polar_decomposition(X):
    """Perform a polar decomposition of a matrix.

    The polar decomposition of a matrix X is a factorization of the form:

    X = P @ Z

    where P is an orthogonal projection matrix and Z is an orthogonal matrix.

    Arguments
    ---------
    X : numpy.ndarray
        Input matrix to decompose. Should be a square matrix.

    Returns
    -------
    dict
        A dictionary containing:
        - "P": the orthogonal projection matrix.
        - "Z": the orthogonal part of the decomposition.
        - "Xtilde": the reconstructed matrix from the decomposition (Xtilde = P @ Z).
    """
    U, d, Vh = np.linalg.svd(X, full_matrices=False) # Vh is V transpose
    # This is the formula for P in a LEFT decomposition (X = PZ)
    P = np.matmul(U, np.matmul(np.diag(d), np.transpose(U)))
    # Z is the orthogonal part, U @ Vh
    Z = np.matmul(U, Vh)

    # Correction to ensure Z is a proper rotation (det(Z) = +1)
    if np.linalg.det(Z) < 0:
        n = X.shape[0]
        reflection_matrix = np.identity(n)
        reflection_matrix[-1, -1] = -1.0 # More robust to change last element
        U_prime = U.copy()
        U_prime[:, -1] *= -1
        Z = U_prime @ Vh

    # The returned Xtilde is P @ Z, consistent with a LEFT decomposition
    return({"P" : P, "Z" : Z, "Xtilde" : np.matmul(P, Z)})



__all__ = ['labels_to_matrix']

import numpy as np

import ants
from ants.decorators import image_method

@image_method
def labels_to_matrix(image, mask, target_labels=None, missing_val=np.nan):
    """
    Convert a labeled image to an n x m binary matrix where n = number of voxels
    and m = number of labels. Only includes values inside the provided mask while
    including background ( image == 0 ) for consistency with timeseries2matrix and
    other image to matrix operations.
    
    ANTsR function: `labels2matrix`

    Arguments
    ---------
    image : ANTsImage
        input label image

    mask : ANTsImage
        defines domain of interest

    target_labels : list/tuple
        defines target regions to be returned.  if the target label does not exist 
        in the input label image, then the matrix will contain a constant value 
        of missing_val (default None) in that row.

    missing_val : scalar
        value to use for missing label values

    Returns
    -------
    ndarray

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16')).resample_image((60,60),1,0)
    >>> mask = ants.get_mask(fi)
    >>> labs = ants.kmeans_segmentation(fi,3)['segmentation']
    >>> labmat = ants.labels_to_matrix(labs, mask)
    """
    if (not ants.is_image(image)) or (not ants.is_image(mask)):
        raise ValueError('image and mask must be ANTsImage types')

    vec = image[mask > 0]
    
    if target_labels is not None:
        the_labels = target_labels
    else:
        the_labels = np.sort(np.unique(vec))

    n_labels = len(the_labels)
    labels = np.zeros((n_labels, len(vec)))

    for i in range(n_labels):
        lab = float(the_labels[i])
        filler = (vec == lab).astype('float')
        if np.sum(filler) == 0:
            filler = np.asarray([np.nan]*len(vec))
        labels[i,:] = filler
    return labels





__all__ = ['RigidRegistration']

from ...registration import interface, apply_transforms


class Registration(object):
    """
    How would it work:

    # Co-registration within-visit
    reg = Registration('Rigid', fixed_image=t1_template, 
                       save_dir=save_dir, save_suffix='_coreg')
    for img in other_imgs:
        reg.fit(img)
    
    for img in [flair_img, t2_img]:
        reg = Registration('Rigid', fixed_image=t1_template,
                           save_dir=save_dir, save_suffix='_coreg')
        reg.fit(img)
    
    # Co-registration across-visit
    reg = Registration('Rigid', fixed_image=t1)
    reg.fit(moving=t1_followup)
    
    # now align all followups with first visit
    for img in [flair_follwup, t2_followup]:
        img_reg = reg.transform(img)
    
    # conversly, align all first visits with followups
    for img in [flair, t2]:
        img_reg = reg.inverse_transform(img)
    """

    def __init__(self, type_of_transform, fixed_image):
        """
        Properties:
            type_of_transform
            fixed_image (template)
            save_dir (where to save outputs)
            save_suffix (what to append to saved outputs)
            save_prefix (what to preppend to saved outputs)
        """
        self.type_of_transform = type_of_transform
        self.fixed_image = fixed_image

    def fit(self, X, y=None):
        """
        X : ANTsImage | string | list of ANTsImage types | list of strings
            images to register to fixed image

        y : string | list of strings
            labels for images
        """
        moving_images = X if isinstance(X, (list,tuple)) else [X]
        moving_labels = y if y is not None else [i for i in range(len(moving_images))]
        fixed_image = self.fixed_image

        self.fwdtransforms_ = {}
        self.invtransforms_ = {}
        self.warpedmovout_ = {}
        self.warpedfixout_ = {}
        
        for moving_image, moving_label in zip(moving_images, moving_labels):
            fit_result = interface.registration(fixed_image,
                                                moving_image,
                                                type_of_transform=self.type_of_transform,
                                                initial_transform=None,
                                                outprefix='',
                                                mask=None,
                                                grad_step=0.2,
                                                flow_sigma=3,
                                                total_sigma=0,
                                                aff_metric='mattes',
                                                aff_sampling=32,
                                                syn_metric='mattes',
                                                syn_sampling=32,
                                                reg_iterations=(40,20,0),
                                                verbose=False)

            self.fwdtransforms_[moving_label] = fit_result['fwdtransforms']
            self.invtransforms_[moving_label] = fit_result['invtransforms']
            self.warpedmovout_[moving_label]  = fit_result['warpedmovout']
            self.warpedfixout_[moving_label]  = fit_result['warpedfixout']

        return self

    def transform(self, X, y=None):
        pass


class RigidRegistration(object):
    """
    Rigid Registration as a Scikit-Learn compatible transform class

    Example
    -------
    >>> import ants
    >>> import ants.extra as extrants
    >>> fi = ants.image_read(ants.get_data('r16'))
    >>> mi = ants.image_read(ants.get_data('r64'))
    >>> regtx = extrants.RigidRegistration()
    >>> regtx.fit(fi, mi)
    >>> mi_r = regtx.transform(mi)
    >>> ants.plot(fi, mi_r.iMath_Canny(1, 2, 4).iMath('MD',1))
    """
    def __init__(self, fixed_image=None):
        self.type_of_transform = 'Rigid'
        self.fixed_image = fixed_image

    def fit(self, moving_image, fixed_image=None):
        if fixed_image is None:
            if self.fixed_image is None:
                raise ValueError('must give fixed_image in fit() or set it in __init__')
            fixed_image = self.fixed_image

        fit_result = interface.registration(fixed_image,
                                            moving_image,
                                            type_of_transform=self.type_of_transform,
                                            initial_transform=None,
                                            outprefix='',
                                            mask=None,
                                            grad_step=0.2,
                                            flow_sigma=3,
                                            total_sigma=0,
                                            aff_metric='mattes',
                                            aff_sampling=32,
                                            syn_metric='mattes',
                                            syn_sampling=32,
                                            reg_iterations=(40,20,0),
                                            verbose=False)
        self._fit_result = fit_result
        self.fwdtransforms_ = fit_result['fwdtransforms']
        self.invtransforms_ = fit_result['invtransforms']
        self.warpedmovout_ = fit_result['warpedmovout']
        self.warpedfiout_ = fit_result['warpedfixout']

    def transform(self, moving_image, fixed_image=None):
        result = apply_transforms(fixed=fixed_image, moving=moving_image,
                                  transformlist=self.fwdtransforms)
        return result

    def inverse_transform(self, moving_image, fixed_image=None):
        result = apply_transforms(fixed=fixed_image, moving=moving_image,
                                  transformlist=self.invtransforms)
        return result





"""
Test ants.learn module

nptest.assert_allclose
self.assertEqual
self.assertTrue
"""

import os
import unittest
from common import run_tests
from tempfile import mktemp

# Prevent displaying figures
import matplotlib as mpl
backend_ =  mpl.get_backend() 
mpl.use("Agg")  

import numpy as np
import numpy.testing as nptest

import ants


class TestModule_plot(unittest.TestCase):

    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img3d = ants.image_read(ants.get_ants_data('mni'))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_plot_example(self):
        filename = mktemp(suffix='.png')
        for img in self.imgs:
            ants.plot(img)
            ants.plot(img, overlay=img*2)
            ants.plot(img, overlay=img*2)
            ants.plot(img, filename=filename)
    
    def test_extra_plot(self):
        img = ants.image_read(ants.get_ants_data('r16'))
        ants.plot(img, overlay=img*2, domain_image_map=ants.image_read(ants.get_data('r64')))

        img = ants.image_read(ants.get_ants_data('r16'))
        ants.plot(img, crop=True)

        img = ants.image_read(ants.get_ants_data('mni'))
        ants.plot(img, overlay=img*2, 
                  domain_image_map=ants.image_read(ants.get_data('mni')).resample_image((4,4,4)))

        img = ants.image_read(ants.get_ants_data('mni'))
        ants.plot(img, overlay=img*2, reorient=True, crop=True)
        
    def test_random(self):
        img = ants.image_read(ants.get_ants_data('r16'))
        img3 = ants.image_read(ants.get_data('r64'))
        img2 = ants.image_read(ants.get_ants_data('mni'))
        imgv = ants.merge_channels([img2])
        
        ants.plot(img2, axis='x', scale=True, ncol=1)
        ants.plot(img2, axis='y', scale=(0.05, 0.95))
        ants.plot(img2, axis='z', slices=[10,20,30], title='Test', cbar=True,
                  cbar_vertical=True)
        ants.plot(img2, cbar=True, cbar_vertical=False)
        ants.plot(img, black_bg=False, title='Test', cbar=True)
        imgx = img2.clone()
        imgx.set_spacing((10,1,1))
        ants.plot(imgx)
        ants.plot(ants.get_ants_data('r16'), overlay=ants.get_data('r64'), blend=True)
        with self.assertRaises(Exception):
            ants.plot(ants.get_ants_data('r16'), overlay=123)
        with self.assertRaises(Exception):
            ants.plot(ants.get_ants_data('r16'), overlay=ants.merge_channels([img,img]))
        ants.plot(ants.from_numpy(np.zeros((100,100))))
        ants.plot(img.clone('unsigned int'))
        
        ants.plot(img, domain_image_map=img3)
        
        with self.assertRaises(Exception):
            ants.plot(123)
        
        
        
        
        
        
        
class TestModule_plot_ortho(unittest.TestCase):

    def setUp(self):
        img3d = ants.image_read(ants.get_ants_data('mni'))
        self.imgs = [img3d]

    def tearDown(self):
        pass

    def test_plot_example(self):
        filename = mktemp(suffix='.png')
        for img in self.imgs:
            ants.plot_ortho(img)
            ants.plot_ortho(img, filename=filename)
            
    def test_plot_extra(self):
        img = ants.image_read(ants.get_ants_data('mni'))
        ants.plot_ortho(img, overlay=img*2, 
                  domain_image_map=ants.image_read(ants.get_data('mni')))

        img = ants.image_read(ants.get_ants_data('mni'))
        ants.plot_ortho(img, overlay=img*2, reorient=True, crop=True)
        
    def test_random_params(self):
        img = ants.image_read(ants.get_ants_data('mni')).resample_image((4,4,4))
        img2 = ants.image_read(ants.get_data('r16'))
        ants.plot_ortho(ants.get_data('mni'), overlay=ants.get_data('mni'))
        with self.assertRaises(Exception):
            ants.plot_ortho(123)
        with self.assertRaises(Exception):
            ants.plot_ortho(img2)
        with self.assertRaises(Exception):
            ants.plot_ortho(img, overlay=img2)
            
        imgx = img.clone()
        imgx.set_spacing((3,3,3))
        ants.plot_ortho(img,overlay=imgx)
        ants.plot_ortho(img.clone('unsigned int'),overlay=img, blend=True)
        
        imgx = img.clone()
        imgx.set_spacing((10,1,1))
        ants.plot_ortho(imgx)
        
        ants.plot_ortho(img, flat=True, title='Test', text='This is a test')
        ants.plot_ortho(img, title='Test', text='This is a test', cbar=True)
        
        with self.assertRaises(Exception):
            ants.plot_ortho(img, domain_image_map=123)
            
        with self.assertRaises(Exception):
            ants.plot_orto(ants.merge_channels([img,img]))
        

class TestModule_plot_ortho_stack(unittest.TestCase):

    def setUp(self):
        self.img = ants.image_read(ants.get_ants_data('mni'))

    def tearDown(self):
        pass

    def test_plot_example(self):
        filename = mktemp(suffix='.png')
        ants.plot_ortho_stack([self.img, self.img])
        ants.plot_ortho_stack([self.img, self.img], filename=filename)
        
    def test_extra_ortho_stack(self):
        img = ants.image_read(ants.get_ants_data('mni'))
        ants.plot_ortho_stack([img, img], overlays=[img*2, img*2],
                  domain_image_map=ants.image_read(ants.get_data('mni')))

        img = ants.image_read(ants.get_ants_data('mni'))
        ants.plot_ortho_stack([img, img], overlays=[img*2, img*2],  reorient=True, crop=True)

class TestModule_plot_hist(unittest.TestCase):

    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img3d = ants.image_read(ants.get_ants_data('mni'))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_plot_example(self):
        filename = mktemp(suffix='.png')
        for img in self.imgs:
            ants.plot_hist(img)

class TestModule_plot_grid(unittest.TestCase):

    def setUp(self):
        mni1 = ants.image_read(ants.get_data('mni'))
        mni2 = mni1.smooth_image(1.)
        mni3 = mni1.smooth_image(2.)
        mni4 = mni1.smooth_image(3.)
        self.images3d = np.asarray([[mni1, mni2],
                                    [mni3, mni4]])
        self.images2d = np.asarray([[mni1.slice_image(2,100), mni2.slice_image(2,100)],
                                    [mni3.slice_image(2,100), mni4.slice_image(2,100)]])

    def tearDown(self):
        pass

    def test_plot_example(self):
        ants.plot_grid(self.images3d, slices=100)
        # should take middle slices if none are given
        ants.plot_grid(self.images3d)
        # should work with 2d images
        ants.plot_grid(self.images2d)
        
    def test_examples(self):
        mni1 = ants.image_read(ants.get_data('mni'))
        mni2 = mni1.smooth_image(1.)
        mni3 = mni1.smooth_image(2.)
        mni4 = mni1.smooth_image(3.)
        images = np.asarray([[mni1, mni2],
                            [mni3, mni4]])
        slices = np.asarray([[100, 100],
                            [100, 100]])
        ants.plot_grid(images=images, slices=slices, title='2x2 Grid')
        images2d = np.asarray([[mni1.slice_image(2,100), mni2.slice_image(2,100)],
                            [mni3.slice_image(2,100), mni4.slice_image(2,100)]])
        ants.plot_grid(images=images2d, title='2x2 Grid Pre-Sliced')
        ants.plot_grid(images.reshape(1,4), slices.reshape(1,4), title='1x4 Grid')
        ants.plot_grid(images.reshape(4,1), slices.reshape(4,1), title='4x1 Grid')

        # Padding between rows and/or columns
        ants.plot_grid(images, slices, cpad=0.02, title='Col Padding')
        ants.plot_grid(images, slices, rpad=0.02, title='Row Padding')
        ants.plot_grid(images, slices, rpad=0.02, cpad=0.02, title='Row and Col Padding')

        # Adding plain row and/or column labels
        ants.plot_grid(images, slices, title='Adding Row Labels', rlabels=['Row #1', 'Row #2'])
        ants.plot_grid(images, slices, title='Adding Col Labels', clabels=['Col #1', 'Col #2'])
        ants.plot_grid(images, slices, title='Row and Col Labels',
                    rlabels=['Row 1', 'Row 2'], clabels=['Col 1', 'Col 2'])

        # Making a publication-quality image
        images = np.asarray([[mni1, mni2, mni2],
                            [mni3, mni4, mni4]])
        slices = np.asarray([[100, 100, 100],
                            [100, 100, 100]])
        axes = np.asarray([[0, 1, 2],
                        [0, 1, 2]])
        ants.plot_grid(images, slices, axes, title='Publication Figures with ANTsPy',
                    tfontsize=20, title_dy=0.03, title_dx=-0.04,
                    rlabels=['Row 1', 'Row 2'],
                    clabels=['Col 1', 'Col 2', 'Col 3'],
                    rfontsize=16, cfontsize=16)


class TestModule_plot_ortho_stack(unittest.TestCase):

    def setUp(self):
        pass
    def tearDown(self):
        pass
    
    def test_random_ortho_stack_params(self):
        img = ants.image_read(ants.get_data('mni')).resample_image((4,4,4))
        img2 = ants.image_read(ants.get_data('r16')).resample_image((4,4))
        
        ants.plot_ortho_stack([ants.get_data('mni'), ants.get_data('mni')])
        ants.plot_ortho_stack([ants.get_data('mni'), ants.get_data('mni')],
                              overlays=[ants.get_data('mni'), ants.get_data('mni')])
        
        with self.assertRaises(Exception):
            ants.plot_ortho_stack([1,2,3])
        with self.assertRaises(Exception):
            ants.plot_ortho_stack([img2,img2])
        with self.assertRaises(Exception):
            ants.plot_ortho_stack([img,img], overlays=[img2,img2])
        with self.assertRaises(Exception):
            ants.plot_ortho_stack([img,img], overlays=[1,2])
            
        imgx = img.clone()
        imgx.set_spacing((2,2,2))
        ants.plot_ortho_stack([img,imgx])
        
        imgx.set_spacing((2,1,1))
        ants.plot_ortho_stack([imgx,img])
        
        ants.plot_ortho_stack([img,img], scale=True, transpose=True,
                              title='Test', colpad=1, rowpad=1,
                              xyz_lines=True)
        ants.plot_ortho_stack([img,img], scale=(0.05,0.95))
            
if __name__ == '__main__':
    run_tests()



"""
Apply anti-alias filter on a binary ANTsImage
"""

__all__ = ['anti_alias']

import ants
from ants.internal import get_lib_fn
from ants.decorators import image_method

@image_method
def anti_alias(image):
    """
    Apply Anti-Alias filter to a binary image
    
    ANTsR function: N/A

    Arguments
    ---------
    image : ANTsImage
        binary image to which anti-aliasing will be applied

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_data('r16'))
    >>> mask = ants.get_mask(img)
    >>> mask_aa = ants.anti_alias(mask)
    >>> ants.plot(mask)
    >>> ants.plot(mask_aa)
    """
    if image.pixeltype != 'unsigned char':
        if image.max() > 255.:
            image = (image - image.max()) / (image.max() - image.min())
        image = image.clone('unsigned char')

    libfn = get_lib_fn('antiAlias%s' % image._libsuffix)
    new_ptr = libfn(image.pointer)
    return ants.from_pointer(new_ptr)




__all__ = [
    "matrix_to_images",
    "images_from_matrix",
    "image_list_to_matrix",
    "images_to_matrix",
    "matrix_from_images",
    "timeseries_to_matrix",
    "matrix_to_timeseries"
]

import os
import json
import numpy as np
import warnings

import ants
from ants.decorators import image_method

@image_method
def matrix_to_timeseries(image, matrix, mask=None):
    """
    converts a matrix to a ND image.

    ANTsR function: `matrix2timeseries`

    Arguments
    ---------

    image: reference ND image

    matrix: matrix to convert to image

    mask: mask image defining voxels of interest


    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> img = ants.make_image( (10,10,10,5 ) )
    >>> mask = ants.ndimage_to_list( img )[0] * 0
    >>> mask[ 4:8, 4:8, 4:8 ] = 1
    >>> mat = ants.timeseries_to_matrix( img, mask = mask )
    >>> img2 = ants.matrix_to_timeseries( img,  mat, mask)
    """

    if mask is None:
        mask = temp[0] * 0 + 1
    temp = matrix_to_images(matrix, mask)
    newImage = ants.list_to_ndimage(image, temp)
    ants.copy_image_info(image, newImage)
    return newImage


def matrix_to_images(data_matrix, mask):
    """
    Unmasks rows of a matrix and writes as images

    ANTsR function: `matrixToImages`

    Arguments
    ---------
    data_matrix : numpy.ndarray
        each row corresponds to an image
        array should have number of columns equal to non-zero voxels in the mask

    mask : ANTsImage
        image containing a binary mask. Rows of the matrix are
        unmasked and written as images. The mask defines the output image space

    Returns
    -------
    list of ANTsImage types

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data('r16'))
    >>> msk = ants.get_mask( img )
    >>> img2 = ants.image_read(ants.get_ants_data('r16'))
    >>> img3 = ants.image_read(ants.get_ants_data('r16'))
    >>> mat = ants.image_list_to_matrix([img,img2,img3], msk )
    >>> ilist = ants.matrix_to_images( mat, msk )
    """

    if data_matrix.ndim > 2:
        data_matrix = data_matrix.reshape(data_matrix.shape[0], -1)

    numimages = len(data_matrix)
    numVoxelsInMatrix = data_matrix.shape[1]
    numVoxelsInMask = (mask >= 0.5).sum()
    if numVoxelsInMask != numVoxelsInMatrix:
        raise ValueError(
            "Num masked voxels %i must match data matrix %i"
            % (numVoxelsInMask, numVoxelsInMatrix)
        )

    imagelist = []
    for i in range(numimages):
        img = mask.clone()
        img[mask >= 0.5] = data_matrix[i, :]
        imagelist.append(img)
    return imagelist


images_from_matrix = matrix_to_images


def images_to_matrix(image_list, mask=None, sigma=None, epsilon=0.5):
    """
    Read images into rows of a matrix, given a mask - much faster for
    large datasets as it is based on C++ implementations.

    ANTsR function: `imagesToMatrix`

    Arguments
    ---------
    image_list : list of ANTsImage types
        images to convert to ndarray

    mask : ANTsImage (optional)
        Mask image, voxels in the mask (>= epsilon) are placed in the matrix. If None,
        the first image in image_list is thresholded at its mean value to create a mask.

    sigma : scaler (optional)
        smoothing factor

    epsilon : scalar
        threshold for mask, values >= epsilon are included in the mask.

    Returns
    -------
    ndarray
        array with a row for each image
        shape = (N_IMAGES, N_VOXELS)

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data('r16'))
    >>> img2 = ants.image_read(ants.get_ants_data('r16'))
    >>> img3 = ants.image_read(ants.get_ants_data('r16'))
    >>> mat = ants.image_list_to_matrix([img,img2,img3])
    """
    if mask is None:
        mask = ants.get_mask(image_list[0])

    num_images = len(image_list)
    mask_thresh = mask.clone() >= epsilon
    mask_arr = mask.numpy() >= epsilon
    num_voxels = np.sum(mask_arr)

    data_matrix = np.empty((num_images, num_voxels))
    do_smooth = sigma is not None
    for i, img in enumerate(image_list):
        if do_smooth:
            img = ants.smooth_image(img, sigma, sigma_in_physical_coordinates=True)
        if np.sum(np.array(img.shape) - np.array(mask_thresh.shape)) != 0:
            img = ants.resample_image_to_target(img, mask_thresh, 2)
        data_matrix[i, :] = img[mask_thresh]
    return data_matrix


image_list_to_matrix = images_to_matrix
matrix_from_images = images_to_matrix

@image_method
def timeseries_to_matrix(image, mask=None):
    """
    Convert a timeseries image into a matrix.

    ANTsR function: `timeseries2matrix`

    Arguments
    ---------
    image : image whose slices we convert to a matrix. E.g. a 3D image of size
           x by y by z will convert to a z by x*y sized matrix

    mask : ANTsImage (optional)
        image containing binary mask. voxels in the mask are placed in the matrix

    Returns
    -------
    ndarray
        array with a row for each image
        shape = (N_IMAGES, N_VOXELS)

    Example
    -------
    >>> import ants
    >>> img = ants.make_image( (10,10,10,5 ) )
    >>> mat = ants.timeseries_to_matrix( img )
    """
    temp = ants.ndimage_to_list(image)
    if mask is None:
        mask = temp[0] * 0 + 1
    return image_list_to_matrix(temp, mask)




__all__ = [
    "create_ants_transform",
    "new_ants_transform",
    "read_transform",
    "write_transform",
    "transform_from_displacement_field",
    "transform_to_displacement_field",
    "fsl2antstransform"
]

import os
import numpy as np

import ants
from ants.internal import get_lib_fn, short_ptype


def new_ants_transform(
    precision="float", dimension=3, transform_type="AffineTransform", parameters=None,
    fixed_parameters=None
):
    """
    Create a new ANTsTransform

    ANTsR function: None

    This is a simplified method for creating an ANTsTransform, mostly used internally.
    See create_ants_transform for more options.

    Example
    -------
    >>> import ants
    >>> tx = ants.new_ants_transform()
    """
    libfn = get_lib_fn(
        "newAntsTransform%s%i" % (short_ptype(precision), dimension)
    )
    itk_tx = libfn(precision, dimension, transform_type)
    ants_tx = ants.ANTsTransform(
        precision=precision,
        dimension=dimension,
        transform_type=transform_type,
        pointer=itk_tx,
    )

    if parameters is not None:
        ants_tx.set_parameters(parameters)

    if fixed_parameters is not None:
        ants_tx.set_fixed_parameters(fixed_parameters)

    return ants_tx


def create_ants_transform(
    transform_type="AffineTransform",
    precision="float",
    dimension=3,
    matrix=None,
    offset=None,
    center=None,
    translation=None,
    parameters=None,
    fixed_parameters=None,
    displacement_field=None,
    supported_types=False,
):
    """
    Create and initialize an ANTsTransform

    ANTsR function: `createAntsrTransform`

    Arguments
    ---------
    transform_type : string
        type of transform(s)

    precision : string
        numerical precision

    dimension : integer
        spatial dimension of transform

    matrix : ndarray
        matrix for linear transforms

    offset : tuple/list
        offset for linear transforms

    center : tuple/list
        center for linear transforms

    translation : tuple/list
        translation for linear transforms

    parameters : ndarray/list
        array of parameters

    fixed_parameters : ndarray/list
        array of fixed parameters

    displacement_field : ANTsImage
        multichannel ANTsImage for non-linear transform

    supported_types : boolean
        flag that returns array of possible transforms types

    Returns
    -------
    ANTsTransform or list of ANTsTransform types

    Example
    -------
    >>> import ants
    >>> translation = (3,4,5)
    >>> tx = ants.create_ants_transform( type='Euler3DTransform', translation=translation )
    """

    def _check_arg(arg, dim=1):
        if arg is None:
            if dim == 1:
                return []
            elif dim == 2:
                return [[]]
        elif isinstance(arg, np.ndarray):
            return arg.tolist()
        elif isinstance(arg, (tuple, list)):
            return list(arg)
        else:
            raise ValueError("Incompatible input argument")

    matrix = _check_arg(matrix, dim=2)
    offset = _check_arg(offset)
    center = _check_arg(center)
    translation = _check_arg(translation)
    parameters = _check_arg(parameters)
    fixed_parameters = _check_arg(fixed_parameters)

    matrix_offset_types = {
        "AffineTransform",
        "CenteredAffineTransform",
        "Euler2DTransform",
        "Euler3DTransform",
        "Rigid3DTransform",
        "Rigid2DTransform",
        "QuaternionRigidTransform",
        "Similarity2DTransform",
        "CenteredSimilarity2DTransform",
        "Similarity3DTransform",
        "CenteredRigid2DTransform",
        "CenteredEuler3DTransform",
    }

    # user_matrix_types = {'Affine','CenteredAffine',
    #                     'Euler', 'CenteredEuler',
    #                     'Rigid', 'CenteredRigid', 'QuaternionRigid',
    #                     'Similarity', 'CenteredSimilarity'}

    if supported_types:
        return set(list(matrix_offset_types) + ["DisplacementFieldTransform"])

    # Check for valid dimension
    if (dimension < 2) or (dimension > 4):
        raise ValueError("Unsupported dimension: %i" % dimension)

    # Check for valid precision
    precision_types = ("float", "double")
    if precision not in precision_types:
        raise ValueError("Unsupported Precision %s" % str(precision))

    # Check for supported transform type
    if (transform_type not in matrix_offset_types) and (
        transform_type != "DisplacementFieldTransform"
    ):
        raise ValueError("Unsupported type %s" % str(transform_type))

    # Check parameters with type
    if transform_type == "Euler3DTransform":
        dimension = 3
    elif transform_type == "Euler2DTransform":
        dimension = 2
    elif transform_type == "Rigid3DTransform":
        dimension = 3
    elif transform_type == "QuaternionRigidTransform":
        dimension = 3
    elif transform_type == "Rigid2DTransform":
        dimension = 2
    elif transform_type == "CenteredRigid2DTransform":
        dimension = 2
    elif transform_type == "CenteredEuler3DTransform":
        dimension = 3
    elif transform_type == "Similarity3DTransform":
        dimension = 3
    elif transform_type == "Similarity2DTransform":
        dimension = 2
    elif transform_type == "CenteredSimilarity2DTransform":
        dimension = 2

    # If displacement field
    if displacement_field is not None:
        # raise ValueError('Displacement field transform not currently supported')
        itk_tx = transform_from_displacement_field(displacement_field)
        return ants.ants_transform(itk_tx)

    # Transforms that derive from itk::MatrixOffsetTransformBase
    libfn = get_lib_fn(
        "matrixOffset%s%i" % (short_ptype(precision), dimension)
    )
    itk_tx = libfn(
        transform_type,
        precision,
        dimension,
        matrix,
        offset,
        center,
        translation,
        parameters,
        fixed_parameters,
    )
    return ants.ANTsTransform(
        precision=precision,
        dimension=dimension,
        transform_type=transform_type,
        pointer=itk_tx,
    )


def transform_from_displacement_field(field):
    """
    Convert deformation field (multiChannel image) to ANTsTransform

    ANTsR function: `antsrTransformFromDisplacementField`

    Arguments
    ---------
    field : ANTsImage
        deformation field as multi-channel ANTsImage

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16') )
    >>> mi = ants.image_read(ants.get_ants_data('r64') )
    >>> fi = ants.resample_image(fi,(60,60),1,0)
    >>> mi = ants.resample_image(mi,(60,60),1,0) # speed up
    >>> mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = ('SyN') )
    >>> vec = ants.image_read( mytx['fwdtransforms'][0] )
    >>> atx = ants.transform_from_displacement_field( vec )
    """
    if not ants.is_image(field):
        raise ValueError("field must be ANTsImage type")
    if field.dimension < 2 or field.dimension > 3:
        raise ValueError("Unsupported displacement field dimension: %i" % field.dimension)
    if field.components != field.dimension:
        raise ValueError("Displacement field must have same number of components as the image dimension")
    libfn = get_lib_fn("antsTransformFromDisplacementField")
    field = field.clone("float")
    txptr = libfn(field.pointer)
    return ants.ANTsTransform(
        precision="float",
        dimension=field.dimension,
        transform_type="DisplacementFieldTransform",
        pointer=txptr,
    )


def transform_to_displacement_field(xfrm, ref):
    """
    Convert displacement field ANTsTransform to displacement field

    ANTsR function: `antsrTransformToDisplacementField`

    Arguments
    ---------
    xfrm : displacement field ANTsTransform
        displacement field ANTsTransform

    ref : ANTs Image

    Returns
    -------
    ANTsVectorImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16') )
    >>> mi = ants.image_read(ants.get_ants_data('r64') )
    >>> fi = ants.resample_image(fi,(60,60),1,0)
    >>> mi = ants.resample_image(mi,(60,60),1,0) # speed up
    >>> mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = ('SyN') )
    >>> vec = ants.image_read( mytx['fwdtransforms'][0] )
    >>> atx = ants.transform_from_displacement_field( vec )
    >>> field = ants.transform_to_displacement_field( atx, fi )
    """

    if not xfrm.type == 'DisplacementFieldTransform':
        raise ValueError("Transform must be of DisplacementFieldTransform type")
    libfn = get_lib_fn("antsTransformToDisplacementField")
    field_ptr = libfn(xfrm.pointer, ref.pointer)
    return ants.from_pointer(field_ptr)


def read_transform(filename, precision="float"):
    """
    Read a transform from file

    ANTsR function: `readAntsrTransform`

    Arguments
    ---------
    filename : string
        filename of transform

    precision : string
        numerical precision of transform

    Returns
    -------
    ANTsTransform

    Example
    -------
    >>> import ants
    >>> tx = ants.new_ants_transform(dimension=2)
    >>> tx.set_parameters((0.9,0,0,1.1,10,11))
    >>> ants.write_transform(tx, '~/desktop/tx.mat')
    >>> tx2 = ants.read_transform('~/desktop/tx.mat')
    """
    filename = os.path.expanduser(filename)
    if not os.path.exists(filename):
        raise ValueError("filename does not exist!")

    if filename.endswith('.nii') or filename.endswith('.nii.gz'):
        return transform_from_displacement_field(ants.image_read(filename))

    libfn1 = get_lib_fn("getTransformDimensionFromFile")
    dimensionUse = libfn1(filename)

    libfn2 = get_lib_fn("getTransformNameFromFile")
    transform_type = libfn2(filename)

    libfn3 = get_lib_fn(
        "readTransform%s%i" % (short_ptype(precision), dimensionUse)
    )
    itk_tx = libfn3(filename, dimensionUse, precision)

    return ants.ANTsTransform(
        precision=precision,
        dimension=dimensionUse,
        transform_type=transform_type,
        pointer=itk_tx,
    )


def write_transform(transform, filename):
    """
    Write ANTsTransform to file

    ANTsR function: `writeAntsrTransform`

    Arguments
    ---------
    transform : ANTsTransform
        transform to save

    filename : string
        filename of transform (file extension is ".mat" for affine transforms)

    Returns
    -------
    N/A

    Example
    -------
    >>> import ants
    >>> tx = ants.new_ants_transform(dimension=2)
    >>> tx.set_parameters((0.9,0,0,1.1,10,11))
    >>> ants.write_transform(tx, '~/desktop/tx.mat')
    >>> tx2 = ants.read_transform('~/desktop/tx.mat')
    """
    if not isinstance(transform, ants.ANTsTransform):
        raise Exception('Only ANTsTransform instances can be written to file. Check that you are not passing in a filepath to a saved transform.')
    filename = os.path.expanduser(filename)
    libfn = get_lib_fn("writeTransform")
    libfn(transform.pointer, filename)

def fsl2antstransform(matrix, reference, moving):
    """
    Convert an FSL linear transform to an antsrTransform
    
    ANTsR function: `fsl2antsrtransform`

    Arguments
    ---------
    matrix : ndarray/list
        4x4 matrix of transform parameters

    reference : ANTsImage
        target image

    moving : ANTsImage
        moving image

    Returns
    -------
    ANTsTransform

    Examples
    --------
    >>> import ants
    >>> import numpy as np
    >>> fslmat = np.zeros((4,4))
    >>> np.fill_diagonal(fslmat, 1)
    >>> img = ants.image_read(ants.get_ants_data('ch2'))
    >>> tx = ants.fsl2antstransform(fslmat, img, img)
    """
    if reference.dimension != 3:
        raise ValueError('reference image must be 3 dimensions')

    if reference.pixeltype != 'float':
        reference = reference.clone('float')
    if moving.pixeltype != 'float':
        moving = moving.clone('float')

    libfn = get_lib_fn('fsl2antstransformF3')
    tx_ptr = libfn(list(matrix), 
                    reference.pointer,
                    moving.pointer,
                    1)

    return ants.ANTsTransform(precision='float', dimension=reference.dimension, 
                             transform_type='AffineTransform', pointer=tx_ptr)




"""
Test ants.deeplearn module

nptest.assert_allclose
self.assertEqual
self.assertTrue
"""

import unittest
from common import run_tests

import numpy as np
import numpy.testing as nptest
import pandas as pd

import ants

class TestModule_crop_image_center(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        crop_size = (64, 67)
        image_cropped = ants.crop_image_center(image, crop_size)
        self.assertEqual(image_cropped.shape[0], crop_size[0])
        self.assertEqual(image_cropped.shape[1], crop_size[1])

class TestModule_pad_image_by_factor(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image_padded = ants.pad_image_by_factor(image, 3)
        self.assertEqual(image_padded.shape[0], 258)
        self.assertEqual(image_padded.shape[1], 258)

class TestModule_pad_or_crop_image_to_size(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        re_size = (64, 237)
        image_cropped = ants.pad_or_crop_image_to_size(image, re_size)
        self.assertEqual(image_cropped.shape[0], re_size[0])
        self.assertEqual(image_cropped.shape[1], re_size[1])

class TestModule_data_augmentation(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image1 = ants.image_read(ants.get_ants_data("r16"))
        image2 = ants.image_read(ants.get_ants_data("r64"))
        data_aug = ants.data_augmentation([[image1], [image2]],
                                          number_of_simulations=5,
                                          reference_image=image1)
        self.assertEqual(len(data_aug['simulated_images']), 5)

class TestModule_randomly_transform_image_data(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image1 = ants.image_read(ants.get_ants_data("r16"))
        image2 = ants.image_read(ants.get_ants_data("r64"))
        data_aug = ants.randomly_transform_image_data(image1,
                                                      [[image1], [image2]],
                                                      number_of_simulations=5)
        self.assertEqual(len(data_aug['simulated_images']), 5)

class TestModule_regression_match_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image1 = ants.image_read(ants.get_ants_data("r16"))
        image2 = ants.image_read(ants.get_ants_data("r64"))
        image_matched = ants.regression_match_image(image1, image2)

class TestModule_extract_image_patches(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        patches = ants.extract_image_patches(image, (8, 16))

class TestModule_histogram_warp_image_intensities(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image_xfrm = ants.histogram_warp_image_intensities(image)

class TestModule_simulate_bias_field(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image_xfrm = ants.simulate_bias_field(image)

class TestModule_one_hot(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        seg = ants.kmeans_segmentation(image, 3)['segmentation']
        one_hot = ants.segmentation_to_one_hot(seg.numpy().astype('int'))        
        one_hot_inv = ants.one_hot_to_segmentation(one_hot, seg)        
        one_hot_c = ants.segmentation_to_one_hot(seg.numpy().astype('int'),
                                                 channel_first_ordering=True)        
        one_hot_c_inv = ants.one_hot_to_segmentation(one_hot_c, seg,
                                                     channel_first_ordering=True)        

if __name__ == "__main__":
    run_tests()



"""
Get local ANTsPy data
"""

__all__ = ['get_ants_data',
           'get_data']

import os
import requests
import tempfile

def get_data(file_id=None, target_file_name=None, antsx_cache_directory=None):
    """
    Get ANTsPy test data file

    ANTsR function: `getANTsRData`

    Arguments
    ---------
    name : string
        name of test image tag to retrieve
        Options:
            - 'r16'
            - 'r27'
            - 'r30'
            - 'r62'
            - 'r64'
            - 'r85'
            - 'ch2'
            - 'mni'
            - 'surf'
            - 'pcasl'
    Returns
    -------
    string
        filepath of test image

    Example
    -------
    >>> import ants
    >>> mnipath = ants.get_ants_data('mni')
    """

    def switch_data(argument):
        switcher = {
            "r16": "https://ndownloader.figshare.com/files/28726512",
            "r27": "https://ndownloader.figshare.com/files/28726515",
            "r30": "https://ndownloader.figshare.com/files/28726518",
            "r62": "https://ndownloader.figshare.com/files/28726521",
            "r64": "https://ndownloader.figshare.com/files/28726524",
            "r85": "https://ndownloader.figshare.com/files/28726527",
            "ch2": "https://ndownloader.figshare.com/files/28726494",
            "mni": "https://ndownloader.figshare.com/files/28726500",
            "surf": "https://ndownloader.figshare.com/files/28726530",
            "pcasl": "http://files.figshare.com/1862041/101_pcasl.nii.gz",
        }
        return(switcher.get(argument, "Invalid argument."))

    if antsx_cache_directory is None:
        antsx_cache_directory = os.path.expanduser('~/.antspy/')
    os.makedirs(antsx_cache_directory, exist_ok=True)

    if os.path.isdir(antsx_cache_directory) == False:
        antsx_cache_directory = tempfile.TemporaryDirectory()

    valid_list = ("r16",
                  "r27",
                  "r30",
                  "r62",
                  "r64",
                  "r85",
                  "ch2",
                  "mni",
                  "surf",
                  "pcasl",
                  "show")

    if file_id == "show" or file_id is None:
       return(valid_list)

    url = switch_data(file_id)

    if target_file_name == None:
        if file_id == "pcasl":
            target_file_name = antsx_cache_directory + "pcasl.nii.gz"
        else:
            extension = ".jpg"
            if file_id == "ch2" or file_id == "mni" or file_id == "surf":
                extension = ".nii.gz"
            if extension == ".jpg":
                target_file_name = antsx_cache_directory + file_id + "slice" + extension
            else:
                target_file_name = antsx_cache_directory + file_id + extension

    target_file_name_path = target_file_name
    if target_file_name == None:
        target_file = tempfile.NamedTemporaryFile(prefix=target_file_name, dir=antsx_cache_directory)
        target_file_name_path = target_file.name
        target_file.close()

    if not os.path.exists(target_file_name_path):
        r = requests.get(url)
        with open(target_file_name_path, 'wb') as f:
            f.write(r.content)

    return(target_file_name_path)

get_ants_data = get_data



from .affine_initializer import affine_initializer
from .apply_transforms import apply_transforms, apply_transforms_to_points
from .average_transform import average_affine_transform, average_affine_transform_no_rigid
from .build_template import build_template
from .compose_displacement_fields import compose_displacement_fields
from .create_jacobian_determinant_image import create_jacobian_determinant_image, deformation_gradient
from .create_warped_grid import create_warped_grid
from .fit_bspline_displacement_field import fit_bspline_displacement_field
from .fit_bspline_object_to_scattered_data import fit_bspline_object_to_scattered_data
from .fit_thin_plate_spline_displacement_field import fit_thin_plate_spline_displacement_field
from .integrate_velocity_field import integrate_velocity_field
from .invert_displacement_field import invert_displacement_field
from .landmark_transforms import fit_transform_to_paired_points, fit_time_varying_transform_to_point_sets
from .registration import registration, motion_correction, label_image_registration
from .simulate_displacement_field import simulate_displacement_field



import os
from tempfile import mktemp

import numpy as np

import ants

__all__ = ['average_images']


def average_images( x, normalize=True, mask=None, imagetype=0, sum_image_threshold=3, return_sum_image=False, verbose=False ):
    """
    average a list of images

    images will be resampled automatically to the largest image space;
    this is not a registration so images should be in the same physical
    space to begin with.

    x : a list containing either filenames or antsImages 

    normalize : boolean

    mask : None or integer; this will perform a masked averaging which can 
        be useful when images have only partial coverage. integer greater 
        than zero will perform morphological closing.

    imagetype : integer
        choose 0/1/2/3 mapping to scalar/vector/tensor/time-series

    sum_image_threshold : integer
        only average regions with overlap greater than or equal to this value

    return_sum_image : boolean
        returns the average and the image that show ROI overlap; primarily for debugging

    verbose : boolean
        will print progress

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> x0=[ ants.get_data('r16'), ants.get_data('r27'), ants.get_data('r62'), ants.get_data('r64') ]
    >>> x1=[]
    >>> for k in range(len(x0)):
    >>>     x1.append( ants.image_read( x0[k] ) )
    >>> avg=ants.average_images(x0)
    >>> avg1=ants.average_images(x1)
    >>> avg2=ants.average_images(x1,mask=0)
	>>> avg3=ants.average_images(x1,mask=1,normalize=True)
    """
    import numpy as np

    def gli( y, normalize=False ):
        if isinstance(y,str):
            y=ants.image_read(y)
        if normalize:
            y=y/y.mean()
        return y

    biggest=0
    biggestind=0
    for k in range( len( x ) ):
        locimg = gli( x[k], False )
        sz=np.prod( locimg.shape )
        if sz > biggest:
            biggest=sz
            biggestind=k

    avg = gli( x[biggestind], False ) * 0
    scl = float( 1.0 / len(x))
    if mask is not None:
        sumimg = gli( x[biggestind], False ) * 0

    for k in range( len( x ) ):
        if verbose and k % 20 == 0:
            print( str(k)+'...', end='',flush=True)
        locimg = gli( x[k], normalize )
        temp = ants.resample_image_to_target( locimg, avg, interp_type='linear', imagetype=imagetype )
        avg = avg + temp
        if mask is not None:
            fgmask = ants.threshold_image(temp,'Otsu',1)
            if mask > 0:
                fgmask = ants.morphology(fgmask,"close",mask)
            sumimg = sumimg + fgmask

    if return_sum_image:
        return avg * scl, sumimg
    if mask is None:
        avg = avg * scl
    else:
        nonzero = sumimg > sum_image_threshold
        tozero = sumimg <= sum_image_threshold
        avg[nonzero] = avg[nonzero] / sumimg[nonzero]
        avg[tozero] = 0
    return avg        




"""
Functions for plotting ants images
"""


__all__ = [
    "plot_hist"
]

import fnmatch
import math
import os
import warnings

from matplotlib import gridspec
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.lines as mlines
import matplotlib.patches as patches
import matplotlib.mlab as mlab
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


import numpy as np
from ants.decorators import image_method

@image_method
def plot_hist(
    image,
    threshold=0.0,
    fit_line=False,
    normfreq=True,
    ## plot label arguments
    title=None,
    grid=True,
    xlabel=None,
    ylabel=None,
    ## other plot arguments
    facecolor="green",
    alpha=0.75,
):
    """
    Plot a histogram from an ANTsImage

    Arguments
    ---------
    image : ANTsImage
        image from which histogram will be created
    """
    img_arr = image.numpy().flatten()
    img_arr = img_arr[np.abs(img_arr) > threshold]

    if normfreq != False:
        normfreq = 1.0 if normfreq == True else normfreq
    n, bins, patches = plt.hist(
        img_arr, 50, facecolor=facecolor, alpha=alpha
    )

    if fit_line:
        # add a 'best fit' line
        y = mlab.normpdf(bins, img_arr.mean(), img_arr.std())
        l = plt.plot(bins, y, "r--", linewidth=1)

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if title is not None:
        plt.title(title)

    plt.grid(grid)
    plt.show()



"""
Test ants.learn module

nptest.assert_allclose
self.assertEqual
self.assertTrue
"""

import os
import unittest
from common import run_tests
from tempfile import mktemp

import numpy as np
import numpy.testing as nptest

import ants

class TestModule_decomposition(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_sparse_decom2_example(self):
        mat = np.random.randn(20, 100)
        mat2 = np.random.randn(20, 90)
        mydecom = ants.sparse_decom2(inmatrix = (mat,mat2), 
                                    sparseness=(0.1,0.3), nvecs=3, 
                                    its=3, perms=10)

    def test_initialize_eigenanatomy_example(self):
        mat = np.random.randn(4,100).astype('float32')
        init = ants.initialize_eigenanatomy(mat)

        img = ants.image_read(ants.get_ants_data('r16'))
        segs = ants.kmeans_segmentation(img, 3)
        init = ants.initialize_eigenanatomy(segs['segmentation'])


    def test_eig_seg_example(self):
        mylist = [ants.image_read(ants.get_ants_data('r16')),
                  ants.image_read(ants.get_ants_data('r27')),
                  ants.image_read(ants.get_ants_data('r85'))]
        myseg = ants.eig_seg(ants.get_mask(mylist[0]), mylist)

        mylist = [ants.image_read(ants.get_ants_data('r16')),
                  ants.image_read(ants.get_ants_data('r27')),
                  ants.image_read(ants.get_ants_data('r85'))]
        myseg = ants.eig_seg(ants.get_mask(mylist[0]), mylist, cthresh=2)

        mylist = [ants.image_read(ants.get_ants_data('r16')),
                  ants.image_read(ants.get_ants_data('r27')),
                  ants.image_read(ants.get_ants_data('r85'))]
        myseg = ants.eig_seg(ants.get_mask(mylist[0]), mylist, apply_segmentation_to_images=True)


if __name__ == '__main__':
    run_tests()


"""
Kelly Kapowski algorithm with computing cortical thickness
"""

__all__ = ['kelly_kapowski']

import ants
from ants.internal import get_lib_fn, get_pointer_string, process_arguments


def kelly_kapowski(s, g, w, its=45, r=0.025, m=1.5, gm_label=2, wm_label=3, **kwargs):
    """
    Compute cortical thickness using the DiReCT algorithm.

    Diffeomorphic registration-based cortical thickness based on probabilistic
    segmentation of an image.  This is an optimization algorithm.


    Arguments
    ---------
    s : ANTsimage
        segmentation image

    g : ANTsImage
        gray matter probability image

    w : ANTsImage
        white matter probability image

    its : integer
        convergence params - controls iterations

    r : scalar
        gradient descent update parameter

    m : scalar
        gradient field smoothing parameter

    gm_label : integer
        label for gray matter in the segmentation image

    wm_label : integer
        label for white matter in the segmentation image

    kwargs : keyword arguments
        anything else, see KellyKapowski help in ANTs

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> img = ants.image_read( ants.get_ants_data('r16') ,2)
    >>> img = ants.resample_image(img, (64,64),1,0)
    >>> mask = ants.get_mask( img )
    >>> segs = ants.kmeans_segmentation( img, k=3, kmask = mask)
    >>> thick = ants.kelly_kapowski(s=segs['segmentation'], g=segs['probabilityimages'][1],
                                    w=segs['probabilityimages'][2], its=45,
                                    r=0.5, m=1)
    """
    if ants.is_image(s):
        s = s.clone('unsigned int')

    d = s.dimension
    outimg = g.clone() * 0.0
    kellargs = {'d': d,
                's': "[{},{},{}]".format(get_pointer_string(s),gm_label,wm_label),
                'g': g,
                'w': w,
                'c': "[{}]".format(its),
                'r': r,
                'm': m,
                'o': outimg}
    for k, v in kwargs.items():
        kellargs[k] = v

    processed_kellargs = process_arguments(kellargs)

    libfn = get_lib_fn('KellyKapowski')
    libfn(processed_kellargs)

    # Check thickness is not still all zeros
    if outimg.sum() == 0.0:
        raise RuntimeError("KellyKapowski failed to compute thickness")

    return outimg






"""
Test ants.registration module

nptest.assert_allclose
self.assertEqual
self.assertTrue


class TestModule_geo_seg(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data('simple'))
        image = ants.n3_bias_field_correction(image, 4)
        image = ants.n3_bias_field_correction(image, 2)
        bmk = ants.get_mask(image)
        segs = ants.kmeans_segmentation(image, 3, bmk)
        priors = segs['probabilityimages']
        seg = ants.geo_seg(image, bmk, priors)


"""


import os
import unittest
from common import run_tests
from tempfile import mktemp

import numpy as np
import numpy.testing as nptest

import ants

class TestModule_atropos(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        # test ANTsPy/ANTsR example
        img = ants.image_read(ants.get_ants_data('r16'))
        img = ants.resample_image(img, (64,64), 1, 0)
        mask = ants.get_mask(img)
        ants.atropos( a = img, m = '[0.2,1x1]', c = '[2,0]',  i = 'kmeans[3]', x = mask )

    def test_multiple_inputs(self):
        img = ants.image_read( ants.get_ants_data("r16"))
        img =  ants.resample_image( img, (64,64), 1, 0 )
        mask = ants.get_mask(img)
        segs1 = ants.atropos( a = img, m = '[0.2,1x1]',
           c = '[2,0]',  i = 'kmeans[3]', x = mask )

        # Use probabilities from k-means seg as priors
        segs2 = ants.atropos( a = img, m = '[0.2,1x1]',
           c = '[2,0]',  i = segs1['probabilityimages'], x = mask )

        # multiple inputs
        feats = [img, ants.iMath(img,"Laplacian"), ants.iMath(img,"Grad") ]
        segs3 = ants.atropos( a = feats, m = '[0.2,1x1]',
           c = '[2,0]',  i = segs1['probabilityimages'], x = mask )


class TestModule_joint_label_fusion(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        ref = ants.image_read( ants.get_ants_data('r16'))
        ref = ants.resample_image(ref, (50,50),1,0)
        ref = ants.iMath(ref,'Normalize')
        mi = ants.image_read( ants.get_ants_data('r27'))
        mi2 = ants.image_read( ants.get_ants_data('r30'))
        mi3 = ants.image_read( ants.get_ants_data('r62'))
        mi4 = ants.image_read( ants.get_ants_data('r64'))
        mi5 = ants.image_read( ants.get_ants_data('r85'))
        refmask = ants.get_mask(ref)
        refmask = ants.iMath(refmask,'ME',2) # just to speed things up
        ilist = [mi,mi2,mi3,mi4,mi5]
        seglist = [None]*len(ilist)
        for i in range(len(ilist)):
            ilist[i] = ants.iMath(ilist[i],'Normalize')
            mytx = ants.registration(fixed=ref , moving=ilist[i] ,
                type_of_transform = ('Affine') )
            mywarpedimage = ants.apply_transforms(fixed=ref,moving=ilist[i],
                    transformlist=mytx['fwdtransforms'])
            ilist[i] = mywarpedimage
            seg = ants.threshold_image(ilist[i],'Otsu', 3)
            seglist[i] = ( seg ) + ants.threshold_image( seg, 1, 3 ).morphology( operation='dilate', radius=3 )

        r = 2
        pp = ants.joint_label_fusion(ref, refmask, ilist, r_search=2,
                    label_list=seglist, rad=[r]*ref.dimension )
        pp = ants.joint_label_fusion(ref,refmask,ilist, r_search=2, rad=2 )

    def test_max_lab_plus_one(self):
        ref = ants.image_read( ants.get_ants_data('r16'))
        ref = ants.resample_image(ref, (50,50),1,0)
        ref = ants.iMath(ref,'Normalize')
        mi = ants.image_read( ants.get_ants_data('r27'))
        mi2 = ants.image_read( ants.get_ants_data('r30'))
        mi3 = ants.image_read( ants.get_ants_data('r62'))
        mi4 = ants.image_read( ants.get_ants_data('r64'))
        mi5 = ants.image_read( ants.get_ants_data('r85'))
        refmask = ants.get_mask(ref)
        refmask = ants.iMath(refmask,'ME',2) # just to speed things up
        ilist = [mi,mi2,mi3,mi4,mi5]
        seglist = [None]*len(ilist)
        for i in range(len(ilist)):
            ilist[i] = ants.iMath(ilist[i],'Normalize')
            mytx = ants.registration(fixed=ref , moving=ilist[i] ,
                type_of_transform = ('Affine') )
            mywarpedimage = ants.apply_transforms(fixed=ref,moving=ilist[i],
                    transformlist=mytx['fwdtransforms'])
            ilist[i] = mywarpedimage
            seg = ants.threshold_image(ilist[i],'Otsu', 3)
            seglist[i] = ( seg ) + ants.threshold_image( seg, 1, 3 ).morphology( operation='dilate', radius=3 )

        r = 2
        pp = ants.joint_label_fusion(ref, refmask, ilist, r_search=2,
                    label_list=seglist, rad=[r]*ref.dimension, max_lab_plus_one=True )
        pp = ants.joint_label_fusion(ref,refmask,ilist, r_search=2, rad=2,
                                     max_lab_plus_one=True)



class TestModule_kelly_kapowski(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        img = ants.image_read( ants.get_ants_data('r16') ,2)
        img = ants.resample_image(img, (64,64),1,0)
        mask = ants.get_mask( img )
        segs = ants.kmeans_segmentation( img, k=3, kmask = mask)
        thick = ants.kelly_kapowski(s=segs['segmentation'], g=segs['probabilityimages'][1],
                                    w=segs['probabilityimages'][2], its=45,
                                    r=0.5, m=1)


class TestModule_kmeans(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.image_read(ants.get_ants_data('r16'))
        fi = ants.n3_bias_field_correction(fi, 2)
        seg = ants.kmeans_segmentation(fi, 3)


class TestModule_label_geometry_measures(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.resample_image(ants.image_read( ants.get_ants_data('r16') ), (140,140),1,0)
        seg = ants.kmeans_segmentation( fi, 3 )['segmentation']
        geom = ants.label_geometry_measures(seg,fi)
        expected_cols = ['Label', 'VolumeInVoxels', 'VolumeInMillimeters',
       'SurfaceAreaInMillimetersSquared', 'Eccentricity', 'Elongation',
       'Roundness', 'Flatness', 'Centroid_x', 'Centroid_y', 'AxesLength_x',
       'AxesLength_y', 'BoundingBoxLower_x', 'BoundingBoxLower_y',
       'BoundingBoxUpper_x', 'BoundingBoxUpper_y', 'MeanIntensity',
       'SigmaIntensity', 'MinIntensity', 'MaxIntensity',
       'IntegratedIntensity']
        for col in expected_cols:
            self.assertTrue(col in geom.columns)
        # Label column should have int type
        self.assertTrue(np.issubdtype(geom['Label'].dtype, np.integer))
        # So should VolumeInVoxels
        self.assertTrue(np.issubdtype(geom['VolumeInVoxels'].dtype, np.integer))
        # Check math of VolumeInMillimeters
        self.assertTrue(np.allclose(geom['VolumeInMillimeters'], geom['VolumeInVoxels'] * np.prod(fi.spacing), atol=1e-6))
        # should be three rows
        self.assertEqual(geom.shape[0], 3)



class TestModule_prior_based_segmentation(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.image_read(ants.get_ants_data('r16'))
        seg = ants.kmeans_segmentation(fi,3)
        mask = ants.threshold_image(seg['segmentation'], 1, 1e15)
        priorseg = ants.prior_based_segmentation(fi, seg['probabilityimages'], mask, 0.25, 0.1, 3)


class TestModule_random(unittest.TestCase):

    def setUp(self):
        pass
    def tearDown(self):
        pass

    def test_fuzzy_cmeans(self):
        image = ants.image_read(ants.get_ants_data('r16'))
        mask = ants.get_mask(image)
        fuzzy = ants.fuzzy_spatial_cmeans_segmentation(image, mask, number_of_clusters=3)

    def test_functional_lung(self):
        image = ants.image_read(ants.get_data("mni")).resample_image((4,4,4))
        mask = image.get_mask()
        seg = ants.functional_lung_segmentation(image, mask, verbose=True,
                                                number_of_iterations=1,
                                                number_of_clusters=2,
                                                number_of_atropos_iterations=1)

    def test_anti_alias(self):
        img = ants.image_read(ants.get_data('r16'))
        mask = ants.get_mask(img)
        mask_aa = ants.anti_alias(mask)

if __name__ == '__main__':
    run_tests()



"""
Timings against numpy/itk/nibabel/ants

To run the timings, first install itk, nibabel, and ants.
Then, run this file `timings_io.py` (e.g. `python timings_io.py`)

On a Macbook Pro
---------------
1 Trial:
NIBABEL TIME: 2.902 seconds
ITK TIME: 4.375 seconds
ANTS TIME: 1.713 seconds

20 Trials:
NIBABEL TIME: 56.121 seconds
ITK TIME: 28.780 seconds
ANTS TIME: 33.601 seconds
"""

import os

import numpy as np
import nibabel as nib
import itk
import ants

import time

def time_nifti_to_numpy(N_TRIALS):
    """
    Times how fast a framework can read a nifti file and convert it to numpy
    """
    img_paths = [ants.get_ants_data('mni')]*10
    
    def test_nibabel():
        for img_path in img_paths:
            array = np.asanyarray(nib.load(img_path).dataobj)

    def test_itk():
        for img_path in img_paths:
            array = itk.GetArrayFromImage(itk.imread(img_path))

    def test_ants():
        for img_path in img_paths:
            array = ants.image_read(img_path, pixeltype='float').numpy()

    nib_start = time.time()
    for i in range(N_TRIALS):
        test_nibabel()
    nib_end = time.time()
    print('NIBABEL TIME: %.3f seconds' % (nib_end-nib_start))

    itk_start = time.time()
    for i in range(N_TRIALS):
        test_itk()
    itk_end = time.time()
    print('ITK TIME: %.3f seconds' % (itk_end-itk_start))

    ants_start = time.time()
    for i in range(N_TRIALS):
        test_ants()
    ants_end = time.time()
    print('ANTS TIME: %.3f seconds' % (ants_end-ants_start))


if __name__ == '__main__':
    time_nifti_to_numpy(N_TRIALS=1) # 
    time_nifti_to_numpy(N_TRIALS=20)




"""
Functions for plotting ants images
"""


__all__ = [
    "plot_grid"
]

import fnmatch
import math
import os
import warnings

from matplotlib import gridspec
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.lines as mlines
import matplotlib.patches as patches
import matplotlib.mlab as mlab
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


import numpy as np


def plot_grid(
    images,
    slices=None,
    axes=2,
    # general figure arguments
    figsize=1.0,
    rpad=0,
    cpad=0,
    vmin=None,
    vmax=None,
    colorbar=True,
    cmap="Greys_r",
    # title arguments
    title=None,
    tfontsize=20,
    title_dx=0,
    title_dy=0,
    # row arguments
    rlabels=None,
    rfontsize=14,
    rfontcolor="white",
    rfacecolor="black",
    # column arguments
    clabels=None,
    cfontsize=14,
    cfontcolor="white",
    cfacecolor="black",
    # save arguments
    filename=None,
    dpi=400,
    transparent=True,
    # other args
    **kwargs
):
    """
    Plot a collection of images in an arbitrarily-defined grid

    Matplotlib named colors: https://matplotlib.org/examples/color/named_colors.html

    Arguments
    ---------
    images : list of ANTsImage types
        image(s) to plot.
        if one image, this image will be used for all grid locations.
        if multiple images, they should be arrange in a list the same
        shape as the `gridsize` argument.

    slices : integer or list of integers
        slice indices to plot
        if one integer, this slice index will be used for all images
        if multiple integers, they should be arranged in a list the same
        shape as the `gridsize` argument

    axes : integer or list of integers
        axis or axes along which to plot image slices
        if one integer, this axis will be used for all images
        if multiple integers, they should be arranged in a list the same
        shape as the `gridsize` argument

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> mni1 = ants.image_read(ants.get_data('mni'))
    >>> mni2 = mni1.smooth_image(1.)
    >>> mni3 = mni1.smooth_image(2.)
    >>> mni4 = mni1.smooth_image(3.)
    >>> images = np.asarray([[mni1, mni2],
    ...                      [mni3, mni4]])
    >>> slices = np.asarray([[100, 100],
    ...                      [100, 100]])
    >>> ants.plot_grid(images=images, slices=slices, title='2x2 Grid')
    >>> images2d = np.asarray([[mni1.slice_image(2,100), mni2.slice_image(2,100)],
    ...                        [mni3.slice_image(2,100), mni4.slice_image(2,100)]])
    >>> ants.plot_grid(images=images2d, title='2x2 Grid Pre-Sliced')
    >>> ants.plot_grid(images.reshape(1,4), slices.reshape(1,4), title='1x4 Grid')
    >>> ants.plot_grid(images.reshape(4,1), slices.reshape(4,1), title='4x1 Grid')

    >>> # Padding between rows and/or columns
    >>> ants.plot_grid(images, slices, cpad=0.02, title='Col Padding')
    >>> ants.plot_grid(images, slices, rpad=0.02, title='Row Padding')
    >>> ants.plot_grid(images, slices, rpad=0.02, cpad=0.02, title='Row and Col Padding')

    >>> # Adding plain row and/or column labels
    >>> ants.plot_grid(images, slices, title='Adding Row Labels', rlabels=['Row #1', 'Row #2'])
    >>> ants.plot_grid(images, slices, title='Adding Col Labels', clabels=['Col #1', 'Col #2'])
    >>> ants.plot_grid(images, slices, title='Row and Col Labels',
                       rlabels=['Row 1', 'Row 2'], clabels=['Col 1', 'Col 2'])

    >>> # Making a publication-quality image
    >>> images = np.asarray([[mni1, mni2, mni2],
    ...                      [mni3, mni4, mni4]])
    >>> slices = np.asarray([[100, 100, 100],
    ...                      [100, 100, 100]])
    >>> axes = np.asarray([[0, 1, 2],
                           [0, 1, 2]])
    >>> ants.plot_grid(images, slices, axes, title='Publication Figures with ANTsPy',
                       tfontsize=20, title_dy=0.03, title_dx=-0.04,
                       rlabels=['Row 1', 'Row 2'],
                       clabels=['Col 1', 'Col 2', 'Col 3'],
                       rfontsize=16, cfontsize=16)
    """

    def mirror_matrix(x):
        return x[::-1, :]

    def rotate270_matrix(x):
        return mirror_matrix(x.T)

    def rotate180_matrix(x):
        return x[::-1, ::-1]

    def rotate90_matrix(x):
        return mirror_matrix(x).T

    def flip_matrix(x):
        return mirror_matrix(rotate180_matrix(x))

    def reorient_slice(x, axis):
        if axis != 1:
            x = rotate90_matrix(x)
        if axis == 1:
            x = rotate90_matrix(x)
        x = mirror_matrix(x)
        return x

    def slice_image(img, axis, idx):
        if axis == 0:
            return img[idx, :, :].numpy()
        elif axis == 1:
            return img[:, idx, :].numpy()
        elif axis == 2:
            return img[:, :, idx].numpy()
        elif axis == -1:
            return img[:, :, idx].numpy()
        elif axis == -2:
            return img[:, idx, :].numpy()
        elif axis == -3:
            return img[idx, :, :].numpy()
        else:
            raise ValueError("axis %i not valid" % axis)

    if isinstance(images, np.ndarray):
        images = images.tolist()
    if not isinstance(images, list):
        raise ValueError("images argument must be of type list")
    if not isinstance(images[0], list):
        images = [images]

    if slices is None:
        one_slice = True
    if isinstance(slices, int):
        one_slice = True
    if isinstance(slices, np.ndarray):
        slices = slices.tolist()
    if isinstance(slices, list):
        one_slice = False
        if not isinstance(slices[0], list):
            slices = [slices]
        nslicerow = len(slices)
        nslicecol = len(slices[0])

    nrow = len(images)
    ncol = len(images[0])

    if rlabels is None:
        rlabels = [None] * nrow
    if clabels is None:
        clabels = [None] * ncol

    if not one_slice:
        if (nrow != nslicerow) or (ncol != nslicecol):
            raise ValueError(
                "`images` arg shape (%i,%i) must equal `slices` arg shape (%i,%i)!"
                % (nrow, ncol, nslicerow, nslicecol)
            )

    fig = plt.figure(figsize=((ncol + 1) * 2.5 * figsize, (nrow + 1) * 2.5 * figsize))

    if title is not None:
        basex = 0.5
        basey = 0.9 if clabels[0] is None else 0.95
        fig.suptitle(title, fontsize=tfontsize, x=basex + title_dx, y=basey + title_dy)

    if (cpad > 0) and (rpad > 0):
        bothgridpad = max(cpad, rpad)
        cpad = 0
        rpad = 0
    else:
        bothgridpad = 0.0

    gs = gridspec.GridSpec(
        nrow,
        ncol,
        wspace=bothgridpad,
        hspace=0.0,
        top=1.0 - 0.5 / (nrow + 1),
        bottom=0.5 / (nrow + 1) + cpad,
        left=0.5 / (ncol + 1) + rpad,
        right=1 - 0.5 / (ncol + 1),
    )

    if isinstance(vmin, (int, float)):
        vmins = [vmin] * nrow
    elif vmin is None:
        vmins = [None] * nrow
    else:
        vmins = vmin

    if isinstance(vmax, (int, float)):
        vmaxs = [vmax] * nrow
    elif vmax is None:
        vmaxs = [None] * nrow
    else:
        vmaxs = vmax

    if isinstance(cmap, str):
        cmaps = [cmap] * nrow
    elif cmap is None:
        cmaps = [None] * nrow
    else:
        cmaps = cmap

    for rowidx, rvmin, rvmax, rcmap in zip(range(nrow), vmins, vmaxs, cmaps):
        for colidx in range(ncol):
            ax = plt.subplot(gs[rowidx, colidx])

            if colidx == 0:
                if rlabels[rowidx] is not None:
                    bottom, height = 0.25, 0.5
                    top = bottom + height
                    # add label text
                    ax.text(
                        -0.07,
                        0.5 * (bottom + top),
                        rlabels[rowidx],
                        horizontalalignment="right",
                        verticalalignment="center",
                        rotation="vertical",
                        transform=ax.transAxes,
                        color=rfontcolor,
                        fontsize=rfontsize,
                    )

                    # add label background
                    extra = 0.3 if rowidx == 0 else 0.0

                    rect = patches.Rectangle(
                        (-0.3, 0),
                        0.3,
                        1.0 + extra,
                        facecolor=rfacecolor,
                        alpha=1.0,
                        transform=ax.transAxes,
                        clip_on=False,
                    )
                    ax.add_patch(rect)

            if rowidx == 0:
                if clabels[colidx] is not None:
                    bottom, height = 0.25, 0.5
                    left, width = 0.25, 0.5
                    right = left + width
                    top = bottom + height
                    ax.text(
                        0.5 * (left + right),
                        0.09 + top + bottom,
                        clabels[colidx],
                        horizontalalignment="center",
                        verticalalignment="center",
                        rotation="horizontal",
                        transform=ax.transAxes,
                        color=cfontcolor,
                        fontsize=cfontsize,
                    )

                    # add label background
                    rect = patches.Rectangle(
                        (0, 1.0),
                        1.0,
                        0.3,
                        facecolor=cfacecolor,
                        alpha=1.0,
                        transform=ax.transAxes,
                        clip_on=False,
                    )
                    ax.add_patch(rect)

            tmpimg = images[rowidx][colidx]
            if isinstance(axes, int):
                tmpaxis = axes
            else:
                tmpaxis = axes[rowidx][colidx]
            
            if tmpimg.dimension == 2:
                tmpslice = tmpimg.numpy()
                tmpslice = reorient_slice(tmpslice, tmpaxis)
            else:
                sliceidx = slices[rowidx][colidx] if not one_slice else slices
                if sliceidx is None:
                    sliceidx = math.ceil(tmpimg.shape[tmpaxis] / 2)
                tmpslice = slice_image(tmpimg, tmpaxis, sliceidx)
                tmpslice = reorient_slice(tmpslice, tmpaxis)
            
            im = ax.imshow(tmpslice, cmap=rcmap, aspect="auto", vmin=rvmin, vmax=rvmax)
            ax.axis("off")

        # A colorbar solution with make_axes_locatable will not allow y-scaling of the colorbar.
        # from mpl_toolkits.axes_grid1 import make_axes_locatable
        # divider = make_axes_locatable(ax)
        # cax = divider.append_axes('right', size='5%', pad=0.05)
        if colorbar:
            axins = inset_axes(ax,
                               width="5%",  # width = 5% of parent_bbox width
                               height="90%",  # height : 50%
                               loc='center left',
                               bbox_to_anchor=(1.03, 0., 1, 1),
                               bbox_transform=ax.transAxes,
                               borderpad=0,
            )
            fig.colorbar(im, cax=axins, orientation='vertical')

    if filename is not None:
        filename = os.path.expanduser(filename)
        plt.savefig(filename, dpi=dpi, transparent=transparent, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.show()






__all__ = ['resample_image',
           'resample_image_to_target']

import os

from ants.decorators import image_method
from ants.internal import get_lib_fn, process_arguments

import ants

@image_method
def resample_image(image, resample_params, use_voxels=False, interp_type=1):
    """
    Resample image by spacing or number of voxels with
    various interpolators. Works with multi-channel images.

    ANTsR function: `resampleImage`

    Arguments
    ---------
    image : ANTsImage
        input image

    resample_params : tuple/list
        vector of size dimension with numeric values

    use_voxels : boolean
        True means interpret resample params as voxel counts

    interp_type : integer
        one of 0 (linear), 1 (nearest neighbor), 2 (gaussian), 3 (windowed sinc), 4 (bspline)

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data("r16"))
    >>> finn = ants.resample_image(fi,(50,60),True,0)
    >>> filin = ants.resample_image(fi,(1.5,1.5),False,1)
    >>> img = ants.image_read( ants.get_ants_data("r16"))
    >>> img = ants.merge_channels([img, img])
    >>> outimg = ants.resample_image(img, (128,128), True)
    """
    if image.components == 1:
        inimage = image.clone('float')
        outimage = image.clone('float')
        rsampar = 'x'.join([str(rp) for rp in resample_params])

        args = [image.dimension, inimage, outimage, rsampar, int(use_voxels), interp_type]
        processed_args = process_arguments(args)
        libfn = get_lib_fn('ResampleImage')
        libfn(processed_args)
        outimage = outimage.clone(image.pixeltype)
        return outimage
    else:
        images = ants.split_channels(image)
        new_images = []
        for image in images:
            inimage = image.clone('float')
            outimage = image.clone('float')
            rsampar = 'x'.join([str(rp) for rp in resample_params])

            args = [image.dimension, inimage, outimage, rsampar, int(use_voxels), interp_type]
            processed_args = process_arguments(args)
            libfn = get_lib_fn('ResampleImage')
            libfn(processed_args)
            outimage = outimage.clone(image.pixeltype)
            new_images.append(outimage)
        outimage = ants.merge_channels(new_images, channels_first=image.channels_first)
        return outimage

@image_method
def resample_image_to_target(image, target, interp_type='linear', imagetype=0, verbose=False, **kwargs):
    """
    Resample image by using another image as target reference.
    This function uses ants.apply_transform with an identity matrix
    to achieve proper resampling.

    ANTsR function: `resampleImageToTarget`

    Arguments
    ---------
    image : ANTsImage
        image to resample

    target : ANTsImage
        image of reference, the output will be in this space and will have the same pixel type.

    interp_type : string
        Choice of interpolator. Supports partial matching.
            linear
            nearestNeighbor
            multiLabel for label images but genericlabel is preferred
            gaussian
            bSpline
            cosineWindowedSinc
            welchWindowedSinc
            hammingWindowedSinc
            lanczosWindowedSinc
            genericLabel use this for label images

    imagetype : integer
        choose 0/1/2/3 mapping to scalar/vector/tensor/time-series

    verbose : boolean
        print command and run verbose application of transform.

    kwargs : keyword arguments
        additional arugment passed to antsApplyTransforms C code

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16'))
    >>> fi2mm = ants.resample_image(fi, (2,2), use_voxels=0, interp_type='linear')
    >>> resampled = ants.resample_image_to_target(fi2mm, fi, verbose=True)
    """
    fixed = target
    moving = image
    compose = None
    transformlist = 'identity'
    interpolator = interp_type

    interpolator_oldoptions = ("linear", "nearestNeighbor", "gaussian", "cosineWindowedSinc", "bSpline")
    if isinstance(interp_type, int):
        interpolator = interpolator_oldoptions[interp_type]

    accepted_interpolators = {"linear", "nearestNeighbor", "multiLabel", "gaussian",
                        "bSpline", "cosineWindowedSinc", "welchWindowedSinc",
                        "hammingWindowedSinc", "lanczosWindowedSinc", "genericLabel"}

    if interpolator not in accepted_interpolators:
        raise ValueError('interpolator not supported - see %s' % accepted_interpolators)

    args = [fixed, moving, transformlist, interpolator]

    if not isinstance(fixed, str):
        if ants.is_image(fixed) and ants.is_image(moving):
            inpixeltype = fixed.pixeltype
            warpedmovout = moving.clone()
            f = fixed.clone('float')
            m = moving.clone('float')
            if (moving.dimension == 4) and (fixed.dimension==3) and (imagetype==0):
                raise ValueError('Set imagetype 3 to transform time series images.')

            wmo = warpedmovout
            mytx = ['-t', 'identity']
            if compose is None:
                args = ['-d', fixed.dimension, '-i', m, '-o', wmo, '-r', f, '-n', interpolator] + mytx

            tfn = '%scomptx.nii.gz' % compose if compose is not None else 'NA'
            if compose is not None:
                mycompo = '[%s,1]' % tfn
                args = ['-d', fixed.dimension, '-i', m, '-o', mycompo, '-r', f, '-n', interpolator] + mytx

            myargs = process_arguments(args)

            myverb = int(verbose)

            processed_args = myargs + ['-z', str(1), '-v', str(myverb), '--float', str(1), '-e', str(imagetype)]
            libfn = get_lib_fn('antsApplyTransforms')
            libfn(processed_args)

            if compose is None:
                return warpedmovout.clone(inpixeltype)
            else:
                if os.path.exists(tfn):
                    return tfn
                else:
                    return None
        else:
            return 1
    else:
        processed_args = myargs + ['-z', str(1), '--float', str(1), '-e', str(imagetype)]
        libfn = get_lib_fn('antsApplyTransforms')
        libfn(processed_args)



import numpy as np
import ants

__all__ = ['convergence_monitoring']

def convergence_monitoring(values, window_size=10):

     if len(values) >= window_size:

         u = np.linspace(0.0, 1.0, num=window_size)
         scattered_data = np.expand_dims(values[-window_size:], axis=-1)
         parametric_data = np.expand_dims(u, axis=-1)
         spacing = 1 / (window_size-1)
         bspline_line = ants.fit_bspline_object_to_scattered_data(scattered_data, parametric_data,
             parametric_domain_origin=[0.0], parametric_domain_spacing=[spacing],
             parametric_domain_size=[window_size], number_of_fitting_levels=1, mesh_size=1,
             spline_order=1)
         bspline_slope = -(bspline_line[1][0] - bspline_line[0][0]) / spacing

         return(bspline_slope)

     else:

         return None





 

__all__ = ['image_mutual_information']


from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def image_mutual_information(image1, image2):
    """
    Compute mutual information between two ANTsImage types

    ANTsR function: `antsImageMutualInformation`
    
    Arguments
    ---------
    image1 : ANTsImage
        image 1

    image2 : ANTsImage
        image 2

    Returns
    -------
    scalar

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data('r16') ).clone('float')
    >>> mi = ants.image_read( ants.get_ants_data('r64') ).clone('float')
    >>> mival = ants.image_mutual_information(fi, mi) # -0.1796141
    """
    if (image1.pixeltype != 'float') or (image2.pixeltype != 'float'):
        raise ValueError('Both images must have float pixeltype')

    if image1.dimension != image2.dimension:
        raise ValueError('Both images must have same dimension')

    libfn = get_lib_fn('antsImageMutualInformation%iD' % image1.dimension)
    return libfn(image1.pointer, image2.pointer)



__all__ = ['from_sitk', 'to_sitk']

import numpy as np
import ants

def from_sitk(sitk_image: "SimpleITK.Image") -> ants.ANTsImage:
    """
    Converts a SimpleITK image into an ANTsPy image. Requires SimpleITK.

    Parameters
    ----------
    sitk_image : SimpleITK.Image
        The SimpleITK image to convert.

    Returns
    -------
    ants_image : ANTsImage
        The converted ANTsPy image.

    Raises
    ------
    ValueError
        If the image dimensionality is unsupported.

    Examples
    --------
    >>> import SimpleITK as sitk
    >>> import ants
    >>> img = sitk.ReadImage('brain.nii.gz')
    >>> ants_img = from_sitk(img)

    """
    try:
        import SimpleITK as sitk
    except ImportError:
        raise ImportError(
            "SimpleITK is required to convert to or from SimpleITK images. Install SimpleITK to use this function."
        )

    ndim = sitk_image.GetDimension()
    if ndim not in (2, 3, 4):
        raise ValueError(f"Unsupported SimpleITK image dimension: {ndim}")

    is_vector = sitk_image.GetNumberOfComponentsPerPixel() > 1

    data = sitk.GetArrayViewFromImage(sitk_image)

    if is_vector:
        # Vector images: SimpleITK gives (z,y,x,components) or (y,x,components)
        spatial_shape = data.shape[:-1]  # drop components
    else:
        spatial_shape = data.shape

    image_dimension = len(spatial_shape)

    if image_dimension == 2:
        direction = np.asarray(sitk_image.GetDirection()).reshape((2, 2))
        spacing = list(sitk_image.GetSpacing())
        origin = list(sitk_image.GetOrigin())
    elif image_dimension == 3:
        direction = np.asarray(sitk_image.GetDirection()).reshape((3, 3))
        spacing = list(sitk_image.GetSpacing())
        origin = list(sitk_image.GetOrigin())
    elif image_dimension == 4:
        direction = np.asarray(sitk_image.GetDirection()).reshape((4, 4))
        spacing = list(sitk_image.GetSpacing())
        origin = list(sitk_image.GetOrigin())
    else:
        raise ValueError(f"Unsupported image dimension: {image_dimension}")

    # Reshape the array properly for ANTsPy
    if is_vector:
        data_reshaped = data.transpose(list(range(image_dimension - 1, -1, -1)) + [image_dimension])
    else:
        data_reshaped = data.transpose(list(range(image_dimension - 1, -1, -1)))

    ants_img: ants.ANTsImage = ants.from_numpy(
        data=data_reshaped,
        origin=origin,
        spacing=spacing,
        direction=direction,
        has_components=is_vector
    )

    return ants_img


def to_sitk(ants_image: ants.ANTsImage) -> "SimpleITK.Image":
    """
    Converts an ANTsPy image into a SimpleITK image. Requires SimpleITK.

    Parameters
    ----------
    ants_image : ANTsImage
        The ANTsPy image to convert.

    Returns
    -------
    sitk_image : SimpleITK.Image
        The converted SimpleITK image.

    Raises
    ------
    ValueError
        If the image dimensionality is unsupported.

    Examples
    --------
    >>> import ants
    >>> ants_img = ants.image_read('brain.nii.gz')
    >>> sitk_img = ants.to_sitk(ants_img)

    """
    try:
        import SimpleITK as sitk
    except ImportError:
        raise ImportError(
            "SimpleITK is required to convert to or from SimpleITK images. Install SimpleITK to use this function."
        )

    if ants_image.dimension not in (2, 3, 4):
        raise ValueError(f"Unsupported ANTsPy image dimension: {ants_image.dimension}")

    data = ants_image.view()
    shape = ants_image.shape
    components = ants_image.components

    is_vector = components > 1

    # Reshape for SimpleITK
    if is_vector:
        # data shape (x,y,z,components) -> sitk expects (z,y,x,components)
        array = data.ravel(order="F").reshape(shape[::-1] + (components,))
    else:
        # data shape (x,y,z) -> sitk expects (z,y,x)
        array = data.ravel(order="F").reshape(shape[::-1])

    sitk_img = sitk.GetImageFromArray(array, isVector=is_vector)

    sitk_img.SetOrigin(ants_image.origin)
    sitk_img.SetSpacing(ants_image.spacing)
    sitk_img.SetDirection(ants_image.direction.flatten())

    return sitk_img




 

__all__ = ['threshold_image']

from ants.decorators import image_method
from ants.internal import get_lib_fn, process_arguments

@image_method
def threshold_image(image, low_thresh=None, high_thresh=None, inval=1, outval=0, binary=True):
    """
    Converts a scalar image into a binary image by thresholding operations

    ANTsR function: `thresholdImage`

    Arguments
    ---------
    image : ANTsImage
        Input image to operate on
    
    low_thresh : scalar (optional)
        Lower edge of threshold window
    
    hight_thresh : scalar (optional)
        Higher edge of threshold window
    
    inval : scalar
        Output value for image voxels in between lothresh and hithresh
    
    outval : scalar
        Output value for image voxels lower than lothresh or higher than hithresh

    binary : boolean
        if true, returns binary thresholded image
        if false, return binary thresholded image multiplied by original image
    
    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> timage = ants.threshold_image(image, 0.5, 1e15)
    """
    if high_thresh is None:
        high_thresh = image.max() + 0.01
    if low_thresh is None:
        low_thresh = image.min() - 0.01
    dim = image.dimension
    outimage = image.clone()
    args = [dim, image, outimage, low_thresh, high_thresh, inval, outval]
    processed_args = process_arguments(args)
    libfn = get_lib_fn('ThresholdImage')
    libfn(processed_args)
    if binary:
        return outimage
    else:
        return outimage*image




"""
Test ants_image.py

nptest.assert_allclose
self.assertEqual
self.assertTrue

"""

import os
import unittest
from common import run_tests

from tempfile import mktemp

import numpy as np
import numpy.testing as nptest

import ants


class TestClass_ANTsTransform(unittest.TestCase):
    """
    Test ants.ANTsImage class
    """
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img3d = ants.image_read(ants.get_ants_data('mni'))
        tx2d = ants.create_ants_transform(precision='float', dimension=2)
        tx3d = ants.create_ants_transform(precision='float', dimension=3)

        self.imgs = [img2d, img3d]
        self.txs = [tx2d, tx3d]
        self.pixeltypes = ['unsigned char', 'unsigned int', 'float']

    def tearDown(self):
        pass

    def test__repr__(self):
        for tx in self.txs:
            r = tx.__repr__()

    def test_get_precision(self):
        for tx in self.txs:
            precision = tx.precision
            self.assertEqual(precision, 'float')

    def test_get_dimension(self):
        tx2d = ants.create_ants_transform(precision='float', dimension=2)
        self.assertEqual(tx2d.dimension, 2)

        tx3d = ants.create_ants_transform(precision='float', dimension=3)
        self.assertEqual(tx3d.dimension, 3)

    def test_get_type(self):
        for tx in self.txs:
            ttype = tx.type
            self.assertEqual(ttype, 'AffineTransform')

    def test_get_pointer(self):
        for tx in self.txs:
            ptr = tx.pointer

    def test_get_parameters(self):
        for tx in self.txs:
            params = tx.parameters

    def test_get_fixed_parameters(self):
        for tx in self.txs:
            fixed_params = tx.fixed_parameters

    def test_invert(self):
        for tx in self.txs:
            txinvert = tx.invert()

    def test_apply(self):
        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = tx.apply(data=(1,2,3), data_type='point')

        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = tx.apply(data=(1,2,3), data_type='vector')

        img = ants.image_read(ants.get_ants_data("r16")).clone('float')
        tx = ants.new_ants_transform(dimension=2)
        tx.set_parameters((0.9,0,0,1.1,10,11))
        img2 = tx.apply(data=img, reference=img, data_type='image')

    def test_apply_to_point(self):
        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = tx.apply_to_point((1,2,3)) # should be (2,4,6)

    def test_apply_to_vector(self):
        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = tx.apply_to_vector((1,2,3)) # should be (2,4,6)

    def test_apply_to_image(self):
        for ptype in self.pixeltypes:
            img = ants.image_read(ants.get_ants_data("r16")).clone(ptype)
            tx = ants.new_ants_transform(dimension=2)
            tx.set_parameters((0.9,0,0,1.1,10,11))
            img2 = tx.apply_to_image(img, img)


class TestModule_ants_transform(unittest.TestCase):

    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img3d = ants.image_read(ants.get_ants_data('mni'))
        tx2d = ants.create_ants_transform(precision='float', dimension=2)
        tx3d = ants.create_ants_transform(precision='float', dimension=3)

        self.imgs = [img2d, img3d]
        self.txs = [tx2d, tx3d]
        self.pixeltypes = ['unsigned char', 'unsigned int', 'float']

    def tearDown(self):
        pass

    def test_get_ants_transform_parameters(self):
        for tx in self.txs:
            params = ants.get_ants_transform_parameters(tx)

    def test_set_ants_transform_parameters(self):
        for tx in self.txs:
            ants.set_ants_transform_parameters(tx, tx.parameters**2)

    def test_get_ants_transform_fixed_parameters(self):
        for tx in self.txs:
            params = ants.get_ants_transform_fixed_parameters(tx)

    def test_set_ants_transform_fixed_parameters(self):
        for tx in self.txs:
            ants.set_ants_transform_fixed_parameters(tx, tx.fixed_parameters**2)

    def test_apply_ants_transform(self):
        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = ants.apply_ants_transform(tx, data=(1,2,3), data_type='point')

        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = ants.apply_ants_transform(tx, data=(1,2,3), data_type='vector')

        img = ants.image_read(ants.get_ants_data("r16")).clone('float')
        tx = ants.new_ants_transform(dimension=2)
        tx.set_parameters((0.9,0,0,1.1,10,11))
        img2 = ants.apply_ants_transform(tx, data=img, reference=img, data_type='image')


    def test_apply_ants_transform_to_point(self):
        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = ants.apply_ants_transform_to_point(tx, (1,2,3)) # should be (2,4,6)

    def test_apply_ants_transform_to_vector(self):
        tx = ants.new_ants_transform()
        params = tx.parameters
        tx.set_parameters(params*2)
        pt2 = ants.apply_ants_transform_to_vector(tx, (1,2,3)) # should be (2,4,6)

    def test_apply_ants_transform_to_image(self):
        img = ants.image_read(ants.get_ants_data("r16")).clone('float')
        tx = ants.new_ants_transform(dimension=2)
        tx.set_parameters((0.9,0,0,1.1,10,11))
        img2 = ants.apply_ants_transform_to_image(tx, img, img)

    def test_apply_ants_transform_to_image_displacement_field(self):
        img = ants.image_read(ants.get_ants_data("r27")).clone('float')
        img2 = ants.image_read(ants.get_ants_data("r16")).clone('float')
        reg = ants.registration(fixed=img, moving=img2, type_of_transform="SyN")
        tra = ants.transform_from_displacement_field(ants.image_read(reg['fwdtransforms'][0]))
        deformed = tra.apply_to_image(img2, reference=img)
        deformed_aat = ants.apply_transforms(img, img2, reg['fwdtransforms'][0], singleprecision=True)
        nptest.assert_allclose(deformed.numpy(), deformed_aat.numpy(), atol=1e-6)

    def test_invert_ants_transform(self):
        img = ants.image_read(ants.get_ants_data("r16")).clone('float')
        tx = ants.new_ants_transform(dimension=2)
        tx.set_parameters((0.9,0,0,1.1,10,11))
        img_transformed = tx.apply_to_image(img, img)
        inv_tx = ants.invert_ants_transform(tx)
        img_orig = inv_tx.apply_to_image(img_transformed, img_transformed)

    def test_compose_ants_transforms(self):
        img = ants.image_read(ants.get_ants_data("r16")).clone('float')
        tx = ants.new_ants_transform(dimension=2)
        tx.set_parameters((0.9,0,0,1.1,10,11))
        inv_tx = tx.invert()
        single_tx = ants.compose_ants_transforms([tx, inv_tx])
        img_orig = single_tx.apply_to_image(img, img)

        # different precisions
        tx1 = ants.new_ants_transform(dimension=2, precision='float')
        tx2 = ants.new_ants_transform(dimension=2, precision='double')
        with self.assertRaises(Exception):
            single_tx = ants.compose_ants_transforms([tx1,tx2])

        # different dimensions
        tx1 = ants.new_ants_transform(dimension=2, precision='float')
        tx2 = ants.new_ants_transform(dimension=3, precision='float')
        with self.assertRaises(Exception):
            single_tx = ants.compose_ants_transforms([tx1,tx2])


    def test_transform_index_to_physical_point(self):
        img = ants.make_image((10,10),np.random.randn(100))
        pt = ants.transform_index_to_physical_point(img, (2,2))

        # image not ANTsImage
        with self.assertRaises(Exception):
            ants.transform_index_to_physical_point(2, (2,2))
        # index not tuple/list
        with self.assertRaises(Exception):
            ants.transform_index_to_physical_point(img,img)
        # index not same size as dimension
        with self.assertRaises(Exception):
            ants.transform_index_to_physical_point(img,[2]*(img.dimension+1))

    def test_transform_physical_point_to_index(self):
        img = ants.make_image((10,10),np.random.randn(100))
        idx = ants.transform_physical_point_to_index(img, (2,2))
        img.set_spacing((2,2))
        idx2 = ants.transform_physical_point_to_index(img, (4,4))

        # image not ANTsImage
        with self.assertRaises(Exception):
            ants.transform_physical_point_to_index(2, (2,2))
        # index not tuple/list
        with self.assertRaises(Exception):
            ants.transform_physical_point_to_index(img,img)
        # index not same size as dimension
        with self.assertRaises(Exception):
            ants.transform_physical_point_to_index(img,[2]*(img.dimension+1))


if __name__ == '__main__':
    run_tests()
